"""AST node classes for YARA rules."""

from yaraast.ast.base import *
from yaraast.ast.conditions import *
from yaraast.ast.expressions import *
from yaraast.ast.meta import *
from yaraast.ast.modules import *
from yaraast.ast.rules import *
from yaraast.ast.strings import *

__all__ = [
    "ASTNode",
    "Condition",
    "DictionaryAccess",
    "Expression",
    "Import",
    "Include",
    "Meta",
    "ModuleReference",
    "Rule",
    "StringDefinition",
    "Tag",
    "YaraFile",
]



"""Visitor pattern implementation for AST traversal."""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, TypeVar, cast

from yaraast.ast.base import ASTNode, YaraFile
from yaraast.ast.comments import Comment, CommentGroup
from yaraast.ast.conditions import (
    AtExpression,
    Condition,
    ForExpression,
    ForOfExpression,
    InExpression,
    OfExpression,
)
from yaraast.ast.expressions import (
    ArrayAccess,
    BinaryExpression,
    BooleanLiteral,
    DoubleLiteral,
    Expression,
    FunctionCall,
    Identifier,
    IntegerLiteral,
    MemberAccess,
    ParenthesesExpression,
    RangeExpression,
    RegexLiteral,
    SetExpression,
    StringCount,
    StringIdentifier,
    StringLength,
    StringLiteral,
    StringOffset,
    StringWildcard,
    UnaryExpression,
)
from yaraast.ast.extern import ExternImport, ExternNamespace, ExternRule, ExternRuleReference
from yaraast.ast.meta import Meta
from yaraast.ast.modifiers import StringModifier
from yaraast.ast.operators import DefinedExpression, StringOperatorExpression
from yaraast.ast.pragmas import InRulePragma, Pragma, PragmaBlock
from yaraast.ast.rules import Import, Include, Rule
from yaraast.ast.simple_nodes import Tag
from yaraast.ast.strings import (
    HexAlternative,
    HexByte,
    HexJump,
    HexString,
    HexToken,
    HexWildcard,
    PlainString,
    RegexString,
    StringDefinition,
)

T = TypeVar("T")


class ASTVisitor[T](ABC):
    """Base visitor class for traversing AST nodes."""

    def visit(self, node: ASTNode) -> T:
        """Visit a node by calling its accept method."""
        return node.accept(self)

    # Base nodes
    @abstractmethod
    def visit_yara_file(self, node: YaraFile) -> T:
        """Visit YaraFile node."""

    @abstractmethod
    def visit_import(self, node: Import) -> T:
        """Visit Import node."""

    @abstractmethod
    def visit_include(self, node: Include) -> T:
        """Visit Include node."""

    @abstractmethod
    def visit_rule(self, node: Rule) -> T:
        """Visit Rule node."""

    @abstractmethod
    def visit_tag(self, node: Tag) -> T:
        """Visit Tag node."""

    # String definitions
    @abstractmethod
    def visit_string_definition(self, node: StringDefinition) -> T:
        """Visit StringDefinition node."""

    @abstractmethod
    def visit_plain_string(self, node: PlainString) -> T:
        """Visit PlainString node."""

    @abstractmethod
    def visit_hex_string(self, node: HexString) -> T:
        """Visit HexString node."""

    @abstractmethod
    def visit_regex_string(self, node: RegexString) -> T:
        """Visit RegexString node."""

    @abstractmethod
    def visit_string_modifier(self, node: StringModifier) -> T:
        """Visit StringModifier node."""

    # Hex tokens
    @abstractmethod
    def visit_hex_token(self, node: HexToken) -> T:
        """Visit HexToken node."""

    @abstractmethod
    def visit_hex_byte(self, node: HexByte) -> T:
        """Visit HexByte node."""

    @abstractmethod
    def visit_hex_wildcard(self, node: HexWildcard) -> T:
        """Visit HexWildcard node."""

    @abstractmethod
    def visit_hex_jump(self, node: HexJump) -> T:
        """Visit HexJump node."""

    @abstractmethod
    def visit_hex_alternative(self, node: HexAlternative) -> T:
        """Visit HexAlternative node."""

    @abstractmethod
    def visit_hex_nibble(self, node) -> T:
        """Visit HexNibble node."""

    # Expressions
    @abstractmethod
    def visit_expression(self, node: Expression) -> T:
        """Visit Expression node."""

    @abstractmethod
    def visit_identifier(self, node: Identifier) -> T:
        """Visit Identifier node."""

    @abstractmethod
    def visit_string_identifier(self, node: StringIdentifier) -> T:
        """Visit StringIdentifier node."""

    @abstractmethod
    def visit_string_wildcard(self, node: StringWildcard) -> T:
        """Visit StringWildcard node."""

    @abstractmethod
    def visit_string_count(self, node: StringCount) -> T:
        """Visit StringCount node."""

    @abstractmethod
    def visit_string_offset(self, node: StringOffset) -> T:
        """Visit StringOffset node."""

    @abstractmethod
    def visit_string_length(self, node: StringLength) -> T:
        """Visit StringLength node."""

    @abstractmethod
    def visit_integer_literal(self, node: IntegerLiteral) -> T:
        """Visit IntegerLiteral node."""

    @abstractmethod
    def visit_double_literal(self, node: DoubleLiteral) -> T:
        """Visit DoubleLiteral node."""

    @abstractmethod
    def visit_string_literal(self, node: StringLiteral) -> T:
        """Visit StringLiteral node."""

    @abstractmethod
    def visit_regex_literal(self, node: RegexLiteral) -> T:
        """Visit RegexLiteral node."""

    @abstractmethod
    def visit_boolean_literal(self, node: BooleanLiteral) -> T:
        """Visit BooleanLiteral node."""

    @abstractmethod
    def visit_binary_expression(self, node: BinaryExpression) -> T:
        """Visit BinaryExpression node."""

    @abstractmethod
    def visit_unary_expression(self, node: UnaryExpression) -> T:
        """Visit UnaryExpression node."""

    @abstractmethod
    def visit_parentheses_expression(self, node: ParenthesesExpression) -> T:
        """Visit ParenthesesExpression node."""

    @abstractmethod
    def visit_set_expression(self, node: SetExpression) -> T:
        """Visit SetExpression node."""

    @abstractmethod
    def visit_range_expression(self, node: RangeExpression) -> T:
        """Visit RangeExpression node."""

    @abstractmethod
    def visit_function_call(self, node: FunctionCall) -> T:
        """Visit FunctionCall node."""

    @abstractmethod
    def visit_array_access(self, node: ArrayAccess) -> T:
        """Visit ArrayAccess node."""

    @abstractmethod
    def visit_member_access(self, node: MemberAccess) -> T:
        """Visit MemberAccess node."""

    # Conditions
    @abstractmethod
    def visit_condition(self, node: Condition) -> T:
        """Visit Condition node."""

    @abstractmethod
    def visit_for_expression(self, node: ForExpression) -> T:
        """Visit ForExpression node."""

    @abstractmethod
    def visit_for_of_expression(self, node: ForOfExpression) -> T:
        """Visit ForOfExpression node."""

    @abstractmethod
    def visit_at_expression(self, node: AtExpression) -> T:
        """Visit AtExpression node."""

    @abstractmethod
    def visit_in_expression(self, node: InExpression) -> T:
        """Visit InExpression node."""

    @abstractmethod
    def visit_of_expression(self, node: OfExpression) -> T:
        """Visit OfExpression node."""

    # Meta
    @abstractmethod
    def visit_meta(self, node: Meta) -> T:
        """Visit Meta node."""

    # Modules
    @abstractmethod
    def visit_module_reference(self, node) -> T:
        """Visit ModuleReference node."""

    @abstractmethod
    def visit_dictionary_access(self, node) -> T:
        """Visit DictionaryAccess node."""

    # Comments
    @abstractmethod
    def visit_comment(self, node: Comment) -> T:
        """Visit Comment node."""

    @abstractmethod
    def visit_comment_group(self, node: CommentGroup) -> T:
        """Visit CommentGroup node."""

    # Operators
    @abstractmethod
    def visit_defined_expression(self, node: DefinedExpression) -> T:
        """Visit DefinedExpression node."""

    @abstractmethod
    def visit_string_operator_expression(
        self,
        node: StringOperatorExpression,
    ) -> T:
        """Visit StringOperatorExpression node."""

    # Extern rules and references
    @abstractmethod
    def visit_extern_rule(self, node: ExternRule) -> T:
        """Visit ExternRule node."""

    @abstractmethod
    def visit_extern_rule_reference(self, node: ExternRuleReference) -> T:
        """Visit ExternRuleReference node."""

    @abstractmethod
    def visit_extern_import(self, node: ExternImport) -> T:
        """Visit ExternImport node."""

    @abstractmethod
    def visit_extern_namespace(self, node: ExternNamespace) -> T:
        """Visit ExternNamespace node."""

    # Pragmas and directives
    @abstractmethod
    def visit_pragma(self, node: Pragma) -> T:
        """Visit Pragma node."""

    @abstractmethod
    def visit_in_rule_pragma(self, node: InRulePragma) -> T:
        """Visit InRulePragma node."""

    @abstractmethod
    def visit_pragma_block(self, node: PragmaBlock) -> T:
        """Visit PragmaBlock node."""


class BaseVisitor(ASTVisitor[T]):
    """Base visitor with default no-op implementations.

    This class provides default implementations for all abstract visitor
    methods that recursively visit child nodes. Subclasses can override
    only the methods they need to customize without implementing every
    single abstract method.
    """

    def visit_yara_file(self, node: YaraFile) -> T:
        """Visit YaraFile node."""
        for imp in node.imports:
            self.visit(imp)
        for inc in node.includes:
            self.visit(inc)
        for rule in node.rules:
            self.visit(rule)
        for extern_rule in node.extern_rules:
            self.visit(extern_rule)
        for extern_import in node.extern_imports:
            self.visit(extern_import)
        for pragma in node.pragmas:
            self.visit(pragma)
        for namespace in node.namespaces:
            self.visit(namespace)
        return cast(T, None)

    def visit_import(self, node: Import) -> T:
        """Visit Import node."""
        return cast(T, None)

    def visit_include(self, node: Include) -> T:
        """Visit Include node."""
        return cast(T, None)

    def visit_rule(self, node: Rule) -> T:
        """Visit Rule node."""
        for tag in node.tags:
            self.visit(tag)
        for string in node.strings:
            self.visit(string)
        if node.condition:
            self.visit(node.condition)
        for pragma in node.pragmas:
            self.visit(pragma)
        return cast(T, None)

    def visit_tag(self, node: Tag) -> T:
        """Visit Tag node."""
        return cast(T, None)

    def visit_string_definition(self, node: StringDefinition) -> T:
        """Visit StringDefinition node."""
        return cast(T, None)

    def visit_plain_string(self, node: PlainString) -> T:
        """Visit PlainString node."""
        for modifier in node.modifiers:
            self.visit(modifier)
        return cast(T, None)

    def visit_hex_string(self, node: HexString) -> T:
        """Visit HexString node."""
        for token in node.tokens:
            self.visit(token)
        for modifier in node.modifiers:
            self.visit(modifier)
        return cast(T, None)

    def visit_regex_string(self, node: RegexString) -> T:
        """Visit RegexString node."""
        for modifier in node.modifiers:
            self.visit(modifier)
        return cast(T, None)

    def visit_string_modifier(self, node: StringModifier) -> T:
        """Visit StringModifier node."""
        return cast(T, None)

    def visit_hex_token(self, node: HexToken) -> T:
        """Visit HexToken node."""
        return cast(T, None)

    def visit_hex_byte(self, node: HexByte) -> T:
        """Visit HexByte node."""
        return cast(T, None)

    def visit_hex_wildcard(self, node: HexWildcard) -> T:
        """Visit HexWildcard node."""
        return cast(T, None)

    def visit_hex_jump(self, node: HexJump) -> T:
        """Visit HexJump node."""
        return cast(T, None)

    def visit_hex_alternative(self, node: HexAlternative) -> T:
        """Visit HexAlternative node."""
        for alternative in node.alternatives:
            for token in alternative:
                self.visit(token)
        return cast(T, None)

    def visit_hex_nibble(self, node) -> T:
        """Visit HexNibble node."""
        return cast(T, None)

    def visit_expression(self, node: Expression) -> T:
        """Visit Expression node."""
        return cast(T, None)

    def visit_identifier(self, node: Identifier) -> T:
        """Visit Identifier node."""
        return cast(T, None)

    def visit_string_identifier(self, node: StringIdentifier) -> T:
        """Visit StringIdentifier node."""
        return cast(T, None)

    def visit_string_wildcard(self, node: StringWildcard) -> T:
        """Visit StringWildcard node."""
        return cast(T, None)

    def visit_string_count(self, node: StringCount) -> T:
        """Visit StringCount node."""
        return cast(T, None)

    def visit_string_offset(self, node: StringOffset) -> T:
        """Visit StringOffset node."""
        if node.index:
            self.visit(node.index)
        return cast(T, None)

    def visit_string_length(self, node: StringLength) -> T:
        """Visit StringLength node."""
        if node.index:
            self.visit(node.index)
        return cast(T, None)

    def visit_integer_literal(self, node: IntegerLiteral) -> T:
        """Visit IntegerLiteral node."""
        return cast(T, None)

    def visit_double_literal(self, node: DoubleLiteral) -> T:
        """Visit DoubleLiteral node."""
        return cast(T, None)

    def visit_string_literal(self, node: StringLiteral) -> T:
        """Visit StringLiteral node."""
        return cast(T, None)

    def visit_regex_literal(self, node: RegexLiteral) -> T:
        """Visit RegexLiteral node."""
        return cast(T, None)

    def visit_boolean_literal(self, node: BooleanLiteral) -> T:
        """Visit BooleanLiteral node."""
        return cast(T, None)

    def visit_binary_expression(self, node: BinaryExpression) -> T:
        """Visit BinaryExpression node."""
        self.visit(node.left)
        self.visit(node.right)
        return cast(T, None)

    def visit_unary_expression(self, node: UnaryExpression) -> T:
        """Visit UnaryExpression node."""
        self.visit(node.operand)
        return cast(T, None)

    def visit_parentheses_expression(self, node: ParenthesesExpression) -> T:
        """Visit ParenthesesExpression node."""
        self.visit(node.expression)
        return cast(T, None)

    def visit_set_expression(self, node: SetExpression) -> T:
        """Visit SetExpression node."""
        for element in node.elements:
            self.visit(element)
        return cast(T, None)

    def visit_range_expression(self, node: RangeExpression) -> T:
        """Visit RangeExpression node."""
        self.visit(node.low)
        self.visit(node.high)
        return cast(T, None)

    def visit_function_call(self, node: FunctionCall) -> T:
        """Visit FunctionCall node."""
        for arg in node.arguments:
            self.visit(arg)
        return cast(T, None)

    def visit_array_access(self, node: ArrayAccess) -> T:
        """Visit ArrayAccess node."""
        self.visit(node.array)
        self.visit(node.index)
        return cast(T, None)

    def visit_member_access(self, node: MemberAccess) -> T:
        """Visit MemberAccess node."""
        self.visit(node.object)
        return cast(T, None)

    def visit_condition(self, node: Condition) -> T:
        """Visit Condition node."""
        return cast(T, None)

    def visit_for_expression(self, node: ForExpression) -> T:
        """Visit ForExpression node."""
        self.visit(node.iterable)
        self.visit(node.body)
        return cast(T, None)

    def visit_for_of_expression(self, node: ForOfExpression) -> T:
        """Visit ForOfExpression node."""
        self.visit(node.string_set)
        if node.condition:
            self.visit(node.condition)
        return cast(T, None)

    def visit_at_expression(self, node: AtExpression) -> T:
        """Visit AtExpression node."""
        self.visit(node.offset)
        return cast(T, None)

    def visit_in_expression(self, node: InExpression) -> T:
        """Visit InExpression node."""
        self.visit(node.range)
        return cast(T, None)

    def visit_of_expression(self, node: OfExpression) -> T:
        """Visit OfExpression node."""
        self.visit(node.quantifier)
        self.visit(node.string_set)
        return cast(T, None)

    def visit_meta(self, node: Meta) -> T:
        """Visit Meta node."""
        return cast(T, None)

    def visit_module_reference(self, node) -> T:
        """Visit ModuleReference node."""
        return cast(T, None)

    def visit_dictionary_access(self, node) -> T:
        """Visit DictionaryAccess node."""
        self.visit(node.object)
        if isinstance(node.key, ASTNode):
            self.visit(node.key)
        return cast(T, None)

    def visit_comment(self, node: Comment) -> T:
        """Visit Comment node."""
        return cast(T, None)

    def visit_comment_group(self, node: CommentGroup) -> T:
        """Visit CommentGroup node."""
        for comment in node.comments:
            self.visit(comment)
        return cast(T, None)

    def visit_defined_expression(self, node: DefinedExpression) -> T:
        """Visit DefinedExpression node."""
        self.visit(node.expression)
        return cast(T, None)

    def visit_string_operator_expression(
        self,
        node: StringOperatorExpression,
    ) -> T:
        """Visit StringOperatorExpression node."""
        self.visit(node.left)
        self.visit(node.right)
        return cast(T, None)

    def visit_extern_rule(self, node: ExternRule) -> T:
        """Visit ExternRule node."""
        return cast(T, None)

    def visit_extern_rule_reference(self, node: ExternRuleReference) -> T:
        """Visit ExternRuleReference node."""
        return cast(T, None)

    def visit_extern_import(self, node: ExternImport) -> T:
        """Visit ExternImport node."""
        return cast(T, None)

    def visit_extern_namespace(self, node: ExternNamespace) -> T:
        """Visit ExternNamespace node."""
        return cast(T, None)

    def visit_pragma(self, node: Pragma) -> T:
        """Visit Pragma node."""
        return cast(T, None)

    def visit_in_rule_pragma(self, node: InRulePragma) -> T:
        """Visit InRulePragma node."""
        self.visit(node.pragma)
        return cast(T, None)

    def visit_pragma_block(self, node: PragmaBlock) -> T:
        """Visit PragmaBlock node."""
        for pragma in node.pragmas:
            self.visit(pragma)
        return cast(T, None)


class ASTTransformer(ASTVisitor[ASTNode]):
    """Base transformer class for modifying AST nodes."""

    def visit_yara_file(self, node: YaraFile) -> YaraFile:
        """Transform YaraFile node."""
        from typing import cast

        imports = [self.visit(imp) for imp in node.imports]
        includes = [self.visit(inc) for inc in node.includes]
        rules = cast("list[Rule]", [self.visit(rule) for rule in node.rules])
        extern_rules = cast(
            "list[ExternRule]",
            [self.visit(r) for r in node.extern_rules],
        )
        extern_imports = cast(
            "list[ExternImport]",
            [self.visit(i) for i in node.extern_imports],
        )
        pragmas = cast(
            "list[Pragma]",
            [self.visit(p) for p in node.pragmas],
        )
        namespaces = [self.visit(namespace) for namespace in node.namespaces]

        return YaraFile(
            imports=imports,
            includes=includes,
            rules=rules,
            extern_rules=extern_rules,
            extern_imports=extern_imports,
            pragmas=pragmas,
            namespaces=namespaces,
        )

    def visit_import(self, node: Import) -> Import:
        """Transform Import node."""
        return Import(module=node.module, location=node.location)

    def visit_include(self, node: Include) -> Include:
        """Transform Include node."""
        return Include(path=node.path, location=node.location)

    def visit_rule(self, node: Rule) -> Rule:
        """Transform Rule node."""
        tags = [self.visit(tag) for tag in node.tags]
        strings = [self.visit(s) for s in node.strings]
        condition = self.visit(node.condition) if node.condition else None
        pragmas = [self.visit(pragma) for pragma in node.pragmas]

        # Handle enhanced meta entries
        meta = node.meta.copy() if isinstance(node.meta, list) else dict(node.meta)

        return Rule(
            name=node.name,
            modifiers=node.modifiers[:],
            tags=tags,
            meta=meta,
            strings=strings,
            condition=condition,
            pragmas=pragmas,
            location=node.location,
        )

    def visit_tag(self, node: Tag) -> Tag:
        """Transform Tag node."""
        return Tag(name=node.name, location=node.location)

    def visit_string_definition(
        self,
        node: StringDefinition,
    ) -> StringDefinition:
        """Transform StringDefinition node."""
        return node  # Base implementation

    def visit_plain_string(self, node: PlainString) -> PlainString:
        """Transform PlainString node."""
        modifiers = [self.visit(mod) for mod in node.modifiers]
        return PlainString(
            identifier=node.identifier,
            value=node.value,
            modifiers=modifiers,
            location=node.location,
        )

    def visit_hex_string(self, node: HexString) -> HexString:
        """Transform HexString node."""
        tokens = [cast("HexToken", self.visit(token)) for token in node.tokens]
        modifiers = [cast("StringModifier", self.visit(mod)) for mod in node.modifiers]
        return HexString(
            identifier=node.identifier,
            tokens=tokens,
            modifiers=modifiers,
            location=node.location,
        )

    def visit_regex_string(self, node: RegexString) -> RegexString:
        """Transform RegexString node."""
        modifiers = [self.visit(mod) for mod in node.modifiers]
        return RegexString(
            identifier=node.identifier,
            regex=node.regex,
            modifiers=modifiers,
            location=node.location,
        )

    def visit_string_modifier(self, node: StringModifier) -> StringModifier:
        """Transform StringModifier node."""
        return StringModifier(name=node.name, value=node.value, location=node.location)

    def visit_hex_token(self, node: HexToken) -> HexToken:
        """Transform HexToken node."""
        return node  # Base implementation

    def visit_hex_byte(self, node: HexByte) -> HexByte:
        """Transform HexByte node."""
        return HexByte(value=node.value, location=node.location)

    def visit_hex_wildcard(self, node: HexWildcard) -> HexWildcard:
        """Transform HexWildcard node."""
        return HexWildcard(location=node.location)

    def visit_hex_jump(self, node: HexJump) -> HexJump:
        """Transform HexJump node."""
        return HexJump(min_jump=node.min_jump, max_jump=node.max_jump)

    def visit_hex_alternative(self, node: HexAlternative) -> HexAlternative:
        """Transform HexAlternative node."""
        alternatives = [[self.visit(token) for token in alt] for alt in node.alternatives]
        return HexAlternative(alternatives=alternatives, location=node.location)

    def visit_hex_nibble(self, node) -> Any:
        """Transform HexNibble node."""
        from yaraast.builder.hex_string_builder import HexNibble

        return HexNibble(high=node.high, value=node.value, location=node.location)

    def visit_expression(self, node: Expression) -> Expression:
        """Transform Expression node."""
        return node  # Base implementation

    def visit_identifier(self, node: Identifier) -> Identifier:
        """Transform Identifier node."""
        # expressions.Identifier doesn't accept location
        from yaraast.ast.expressions import Identifier as ExprIdentifier

        return ExprIdentifier(name=node.name)

    def visit_string_identifier(
        self,
        node: StringIdentifier,
    ) -> StringIdentifier:
        """Transform StringIdentifier node."""
        return StringIdentifier(name=node.name)

    def visit_string_wildcard(
        self,
        node: StringWildcard,
    ) -> StringWildcard:
        """Transform StringWildcard node."""
        return StringWildcard(pattern=node.pattern)

    def visit_string_count(self, node: StringCount) -> StringCount:
        """Transform StringCount node."""
        return StringCount(string_id=node.string_id)

    def visit_string_offset(self, node: StringOffset) -> StringOffset:
        """Transform StringOffset node."""
        index = cast("Expression", self.visit(node.index)) if node.index else None
        return StringOffset(string_id=node.string_id, index=index)

    def visit_string_length(self, node: StringLength) -> StringLength:
        """Transform StringLength node."""
        index = cast("Expression", self.visit(node.index)) if node.index else None
        return StringLength(string_id=node.string_id, index=index)

    def visit_integer_literal(self, node: IntegerLiteral) -> IntegerLiteral:
        """Transform IntegerLiteral node."""
        return IntegerLiteral(value=node.value)

    def visit_double_literal(self, node: DoubleLiteral) -> DoubleLiteral:
        """Transform DoubleLiteral node."""
        return DoubleLiteral(value=node.value)

    def visit_string_literal(self, node: StringLiteral) -> StringLiteral:
        """Transform StringLiteral node."""
        return StringLiteral(value=node.value)

    def visit_regex_literal(self, node: RegexLiteral) -> RegexLiteral:
        """Transform RegexLiteral node."""
        return RegexLiteral(pattern=node.pattern, modifiers=node.modifiers)

    def visit_boolean_literal(self, node: BooleanLiteral) -> BooleanLiteral:
        """Transform BooleanLiteral node."""
        return BooleanLiteral(value=node.value)

    def visit_binary_expression(
        self,
        node: BinaryExpression,
    ) -> BinaryExpression:
        """Transform BinaryExpression node."""
        from typing import cast

        left = cast("Expression", self.visit(node.left))
        right = cast("Expression", self.visit(node.right))
        # BinaryExpression doesn't accept location parameter
        return BinaryExpression(left=left, operator=node.operator, right=right)

    def visit_unary_expression(self, node: UnaryExpression) -> UnaryExpression:
        """Transform UnaryExpression node."""
        operand = cast("Expression", self.visit(node.operand))
        return UnaryExpression(operator=node.operator, operand=operand)

    def visit_parentheses_expression(
        self,
        node: ParenthesesExpression,
    ) -> ParenthesesExpression:
        """Transform ParenthesesExpression node."""
        from typing import cast

        expression = cast("Expression", self.visit(node.expression))
        return ParenthesesExpression(expression=expression)

    def visit_set_expression(self, node: SetExpression) -> SetExpression:
        """Transform SetExpression node."""
        elements = [self.visit(elem) for elem in node.elements]
        return SetExpression(elements=elements, location=node.location)

    def visit_range_expression(self, node: RangeExpression) -> RangeExpression:
        """Transform RangeExpression node."""
        from typing import cast

        low = cast("Expression", self.visit(node.low))
        high = cast("Expression", self.visit(node.high))
        result = RangeExpression(low=low, high=high)
        result.location = node.location
        return result

    def visit_function_call(self, node: FunctionCall) -> FunctionCall:
        """Transform FunctionCall node."""
        arguments = [cast("Expression", self.visit(arg)) for arg in node.arguments]
        return FunctionCall(function=node.function, arguments=arguments)

    def visit_array_access(self, node: ArrayAccess) -> ArrayAccess:
        """Transform ArrayAccess node."""
        array = self.visit(node.array)
        index = self.visit(node.index)
        return ArrayAccess(array=array, index=index, location=node.location)

    def visit_member_access(self, node: MemberAccess) -> MemberAccess:
        """Transform MemberAccess node."""
        obj = self.visit(node.object)
        # Ensure obj is an Expression
        if not isinstance(obj, Expression):
            obj = cast("Expression", obj)
        return MemberAccess(object=obj, member=node.member)

    def visit_condition(self, node: Condition) -> Condition:
        """Transform Condition node."""
        return node  # Base implementation

    def visit_for_expression(self, node: ForExpression) -> ForExpression:
        """Transform ForExpression node."""
        iterable = self.visit(node.iterable)
        body = self.visit(node.body)
        return ForExpression(
            quantifier=node.quantifier,
            variable=node.variable,
            iterable=iterable,
            body=body,
            location=node.location,
        )

    def visit_for_of_expression(
        self,
        node: ForOfExpression,
    ) -> ForOfExpression:
        """Transform ForOfExpression node."""
        string_set = self.visit(node.string_set)
        condition = self.visit(node.condition) if node.condition else None
        return ForOfExpression(
            quantifier=node.quantifier,
            string_set=string_set,
            condition=condition,
            location=node.location,
        )

    def visit_at_expression(self, node: AtExpression) -> AtExpression:
        """Transform AtExpression node."""
        offset = cast("Expression", self.visit(node.offset))
        return AtExpression(string_id=node.string_id, offset=offset)

    def visit_in_expression(self, node: InExpression) -> InExpression:
        """Transform InExpression node."""
        range_expr = cast("Expression", self.visit(node.range))
        return InExpression(string_id=node.string_id, range=range_expr)

    def visit_of_expression(self, node: OfExpression) -> OfExpression:
        """Transform OfExpression node."""
        quantifier = cast("Expression", self.visit(node.quantifier))
        string_set = cast("Expression", self.visit(node.string_set))
        return OfExpression(quantifier=quantifier, string_set=string_set)

    def visit_meta(self, node: Meta) -> Meta:
        """Transform Meta node."""
        return Meta(key=node.key, value=node.value, location=node.location)

    def visit_module_reference(self, node) -> Any:
        """Transform ModuleReference node."""
        from yaraast.ast.modules import ModuleReference

        return ModuleReference(module=node.module, location=node.location)

    def visit_dictionary_access(self, node) -> Any:
        """Transform DictionaryAccess node."""
        from yaraast.ast.modules import DictionaryAccess

        obj = self.visit(node.object)
        key = self.visit(node.key) if isinstance(node.key, Expression) else node.key
        return DictionaryAccess(object=obj, key=key, location=node.location)

    def visit_comment(self, node: Comment) -> Comment:
        """Transform Comment node."""
        return Comment(text=node.text, is_multiline=node.is_multiline)

    def visit_comment_group(self, node: CommentGroup) -> CommentGroup:
        """Transform CommentGroup node."""
        comments = [self.visit(comment) for comment in node.comments]
        return CommentGroup(comments=comments)

    def visit_defined_expression(
        self,
        node: DefinedExpression,
    ) -> DefinedExpression:
        """Transform DefinedExpression node."""
        expression = self.visit(node.expression)
        return DefinedExpression(expression=expression)

    def visit_string_operator_expression(
        self,
        node: StringOperatorExpression,
    ) -> StringOperatorExpression:
        """Transform StringOperatorExpression node."""
        left = self.visit(node.left)
        right = self.visit(node.right)
        return StringOperatorExpression(left=left, operator=node.operator, right=right)

    # Extern rules and references
    def visit_extern_rule(self, node: ExternRule) -> ExternRule:
        """Transform ExternRule node."""
        return ExternRule(
            name=node.name,
            modifiers=node.modifiers[:],
            namespace=node.namespace,
            location=node.location,
        )

    def visit_extern_rule_reference(
        self,
        node: ExternRuleReference,
    ) -> ExternRuleReference:
        """Transform ExternRuleReference node."""
        return ExternRuleReference(
            rule_name=node.rule_name,
            namespace=node.namespace,
            location=node.location,
        )

    def visit_extern_import(self, node: ExternImport) -> ExternImport:
        """Transform ExternImport node."""
        return ExternImport(
            module_path=node.module_path,
            alias=node.alias,
            namespace=node.namespace,
            location=node.location,
        )

    def visit_extern_namespace(self, node: ExternNamespace) -> ExternNamespace:
        """Transform ExternNamespace node."""
        return ExternNamespace(name=node.name, location=node.location)

    # Pragmas and directives
    def visit_pragma(self, node: Pragma) -> Pragma:
        """Transform Pragma node."""
        return Pragma(
            pragma_type=node.pragma_type,
            name=node.name,
            arguments=node.arguments[:],
            scope=node.scope,
            location=node.location,
        )

    def visit_in_rule_pragma(self, node: InRulePragma) -> InRulePragma:
        """Transform InRulePragma node."""
        pragma = cast("Pragma", self.visit(node.pragma))
        return InRulePragma(pragma=pragma, position=node.position)

    def visit_pragma_block(self, node: PragmaBlock) -> PragmaBlock:
        """Transform PragmaBlock node."""
        pragmas = [self.visit(pragma) for pragma in node.pragmas]
        return PragmaBlock(pragmas=pragmas, scope=node.scope, location=node.location)



"""YARA-L specific token types."""

from enum import Enum, auto


class YaraLTokenType(Enum):
    """Token types specific to YARA-L."""

    # YARA-L specific sections
    EVENTS = auto()
    MATCH = auto()
    OUTCOME = auto()
    OPTIONS = auto()

    # Time-related tokens
    OVER = auto()
    BEFORE = auto()
    AFTER = auto()
    WITHIN = auto()

    # Time units
    SECONDS = auto()
    MINUTES = auto()
    HOURS = auto()
    DAYS = auto()

    # Aggregation functions
    COUNT = auto()
    COUNT_DISTINCT = auto()
    SUM = auto()
    MIN = auto()
    MAX = auto()
    AVG = auto()
    ARRAY = auto()
    ARRAY_DISTINCT = auto()
    EARLIEST = auto()
    LATEST = auto()

    # UDM fields prefixes
    METADATA = auto()
    PRINCIPAL = auto()
    TARGET = auto()
    NETWORK = auto()
    SECURITY_RESULT = auto()
    UDM = auto()
    ADDITIONAL = auto()

    # YARA-L operators
    CIDR = auto()
    REGEX = auto()
    IS = auto()
    NULL = auto()

    # Reference lists
    REFERENCE_LIST = auto()  # %list_name%

    # Event variables
    EVENT_VAR = auto()  # $e, $e1, $e2, etc.

    # Field placeholders
    PLACEHOLDER = auto()  # %field%

    # Special
    NOCASE = auto()
    IF = auto()
    ELSE = auto()
    BY = auto()
    EVERY = auto()

    # Symbols specific to YARA-L
    ARROW = auto()  # ->
    DOUBLE_COLON = auto()  # ::

    # Literal types
    TIME_LITERAL = auto()  # 5m, 1h, 7d

    # EOF
    EOF = auto()



"""Fluent builder for hex strings."""

from __future__ import annotations

import builtins
import contextlib
from typing import Self

from yaraast.ast.strings import HexAlternative, HexByte, HexJump, HexNibble, HexToken, HexWildcard


class HexStringBuilder:
    """Fluent builder for constructing hex strings."""

    def __init__(self, identifier: str | None = None) -> None:
        self._tokens: list[HexToken] = []
        self.identifier = identifier

    def byte(self, value: int) -> Self:
        """Add a single byte (alias for add)."""
        return self.add(value)

    def add(self, value: int | str | HexToken) -> Self:
        """Add a hex byte or token."""
        if isinstance(value, int):
            if 0 <= value <= 255:
                self._tokens.append(HexByte(value=value))
            else:
                msg = f"Byte value must be 0-255, got {value}"
                raise ValueError(msg)
        elif isinstance(value, str):
            # Parse hex string
            hex_val = value.upper().replace("0X", "")
            if len(hex_val) == 2:
                try:
                    byte_val = int(hex_val, 16)
                    self._tokens.append(HexByte(value=byte_val))
                except ValueError:
                    msg = f"Invalid hex value: {value}"
                    raise ValueError(msg) from None
            else:
                msg = f"Hex value must be 2 characters, got {value}"
                raise ValueError(msg)
        elif isinstance(value, HexToken):
            self._tokens.append(value)
        else:
            msg = f"Invalid type for hex value: {type(value)}"
            raise TypeError(msg)

        return self

    def add_bytes(self, *values: int | str) -> Self:
        """Add multiple hex bytes."""
        for value in values:
            self.add(value)
        return self

    def wildcard(self, count: int = 1) -> Self:
        """Add wildcard bytes (??)."""
        for _ in range(count):
            self._tokens.append(HexWildcard())
        return self

    def nibble(self, value: str) -> Self:
        """Add a nibble pattern like 'F?', '?F', etc."""
        if len(value) != 2:
            msg = "Nibble must be 2 characters"
            raise ValueError(msg)

        if value[0] == "?" and value[1] != "?":
            # ?X pattern - low nibble
            try:
                nibble_val = int(value[1], 16)
                self._tokens.append(HexNibble(high=False, value=nibble_val))
            except ValueError:
                msg = f"Invalid nibble pattern: {value}"
                raise ValueError(msg) from None
        elif value[0] != "?" and value[1] == "?":
            # X? pattern - high nibble
            try:
                nibble_val = int(value[0], 16)
                self._tokens.append(HexNibble(high=True, value=nibble_val))
            except ValueError:
                msg = f"Invalid nibble pattern: {value}"
                raise ValueError(msg) from None
        else:
            msg = f"Invalid nibble pattern: {value}"
            raise ValueError(msg)

        return self

    def jump(self, min_jump: int | None = None, max_jump: int | None = None) -> Self:
        """Add a jump [n-m]."""
        self._tokens.append(HexJump(min_jump=min_jump, max_jump=max_jump))
        return self

    def jump_exact(self, count: int) -> Self:
        """Add an exact jump [n]."""
        return self.jump(count, count)

    def jump_varying(self, min_val: int, max_val: int) -> Self:
        """Add a varying jump [min-max]."""
        return self.jump(min_val, max_val)

    def jump_up_to(self, max_val: int) -> Self:
        """Add a jump up to max [-max]."""
        return self.jump(None, max_val)

    def jump_at_least(self, min_val: int) -> Self:
        """Add a jump at least min [min-]."""
        return self.jump(min_val, None)

    def jump_any(self) -> Self:
        """Add an unlimited jump [-]."""
        return self.jump(None, None)

    def alternative(self, *alternatives: list[int | str | HexStringBuilder]) -> Self:
        """Add an alternative group (a|b|c)."""
        alt_tokens = []

        for alt in alternatives:
            if isinstance(alt, list):
                # List of values
                tokens = []
                for val in alt:
                    if isinstance(val, int):
                        tokens.append(HexByte(value=val))
                    elif isinstance(val, str):
                        byte_val = int(val, 16)
                        tokens.append(HexByte(value=byte_val))
                    else:
                        msg = f"Invalid alternative value type: {type(val)}"
                        raise TypeError(msg)
                alt_tokens.append(tokens)
            elif isinstance(alt, HexStringBuilder):
                # Nested builder
                alt_tokens.append(alt.build())
            else:
                msg = f"Invalid alternative type: {type(alt)}"
                raise TypeError(msg)

        self._tokens.append(HexAlternative(alternatives=alt_tokens))
        return self

    def group(self, builder_func) -> Self:
        """Add a group using a builder function."""
        inner_builder = HexStringBuilder()
        builder_func(inner_builder)
        self._tokens.extend(inner_builder.build())
        return self

    def pattern(self, pattern: str) -> Self:
        """Add tokens from a pattern string like 'FF ?? [2-4] (AA|BB)'."""
        # This is a simplified parser - in production would need full parser
        parts = pattern.split()

        for part in parts:
            self._process_pattern_part(part)

        return self

    def _process_pattern_part(self, part: str) -> None:
        """Process a single pattern part."""
        if part == "??":
            self.wildcard()
        elif self._is_jump_pattern(part):
            self._process_jump_pattern(part)
        elif self._is_nibble_pattern(part):
            self.nibble(part)
        elif self._is_hex_byte(part):
            self._add_hex_byte_safely(part)

    def _is_jump_pattern(self, part: str) -> bool:
        """Check if part is a jump pattern like [2-4]."""
        return part.startswith("[") and part.endswith("]")

    def _is_nibble_pattern(self, part: str) -> bool:
        """Check if part is a nibble pattern like A?."""
        return len(part) == 2 and "?" in part

    def _is_hex_byte(self, part: str) -> bool:
        """Check if part is a regular hex byte."""
        return len(part) == 2

    def _process_jump_pattern(self, part: str) -> None:
        """Process jump pattern like [2-4]."""
        jump_str = part[1:-1]
        if "-" in jump_str:
            parts = jump_str.split("-")
            min_j = int(parts[0]) if parts[0] else None
            max_j = int(parts[1]) if parts[1] else None
            self.jump(min_j, max_j)
        else:
            self.jump_exact(int(jump_str))

    def _add_hex_byte_safely(self, part: str) -> None:
        """Add hex byte with error suppression."""
        with contextlib.suppress(builtins.BaseException):
            self.add(part)

    def build(self) -> list[HexToken]:
        """Build the list of hex tokens."""
        return self._tokens

    @staticmethod
    def from_bytes(data: bytes) -> HexStringBuilder:
        """Create builder from raw bytes."""
        builder = HexStringBuilder()
        for byte in data:
            builder.add(byte)
        return builder

    @staticmethod
    def from_hex_string(hex_str: str) -> HexStringBuilder:
        """Create builder from hex string."""
        builder = HexStringBuilder()
        hex_str = hex_str.replace(" ", "").upper()

        for i in range(0, len(hex_str), 2):
            if i + 1 < len(hex_str):
                builder.add(hex_str[i : i + 2])

        return builder



"""YARA-L specific CLI commands."""

from pathlib import Path

import click

from yaraast.yaral.enhanced_parser import EnhancedYaraLParser
from yaraast.yaral.generator import YaraLGenerator
from yaraast.yaral.optimizer import YaraLOptimizer
from yaraast.yaral.parser import YaraLParser
from yaraast.yaral.validator import YaraLValidator


def _display_validation_results(file, ast, errors, warnings, strict, output_json):
    """Display validation results in JSON or text format."""
    if output_json:
        _display_json_results(file, errors, warnings, strict)
    else:
        _display_text_results(ast, errors, warnings, strict)


def _display_json_results(file, errors, warnings, strict):
    """Display validation results in JSON format."""
    import json

    result = {
        "file": file,
        "errors": [
            {"severity": e.severity, "message": e.message, "section": e.section} for e in errors
        ],
        "warnings": [
            {"severity": w.severity, "message": w.message, "section": w.section} for w in warnings
        ],
        "valid": len(errors) == 0 and (not strict or len(warnings) == 0),
    }
    click.echo(json.dumps(result, indent=2))


def _display_text_results(ast, errors, warnings, strict):
    """Display validation results in text format."""
    # Display errors
    if errors:
        click.echo("❌ Validation Errors:", err=True)
        for error in errors:
            click.echo(f"  • {error}", err=True)

    # Display warnings
    if warnings:
        click.echo("⚠️  Validation Warnings:")
        for warning in warnings:
            click.echo(f"  • {warning}")

    # Summary
    if not errors and (not strict or not warnings):
        click.echo(f"✅ YARA-L file is valid ({len(ast.rules)} rules)")
    else:
        click.echo(
            f"❌ YARA-L file has {len(errors)} errors and {len(warnings)} warnings",
            err=True,
        )
        if strict and warnings:
            raise click.Abort() from None


def _format_yaral_code(code: str) -> str:
    """Format YARA-L code with proper indentation."""
    lines = code.split("\n")
    formatted_lines = []
    indent = 0

    section_keywords = [
        "rule",
        "meta",
        "events",
        "match",
        "condition",
        "outcome",
        "options",
    ]

    for line in lines:
        stripped = line.strip()
        if not stripped:
            formatted_lines.append("")
            continue

        formatted_line, indent = _format_line(stripped, indent, section_keywords)
        formatted_lines.append(formatted_line)

    return "\n".join(formatted_lines)


def _format_line(stripped: str, indent: int, section_keywords: list) -> tuple[str, int]:
    """Format a single line and return the formatted line and new indent level."""
    if stripped.endswith("{"):
        return ("  " * indent + stripped, indent + 1)
    if stripped.startswith("}"):
        new_indent = max(0, indent - 1)
        return ("  " * new_indent + stripped, new_indent)
    if stripped.endswith(":"):
        line = "  " * indent + stripped
        new_indent = indent + 1 if not stripped.startswith("rule") else indent
        return (line, new_indent)
    if stripped and not any(stripped.startswith(s) for s in section_keywords):
        return ("  " * max(1, indent) + stripped, indent)

    new_indent = 1 if stripped in section_keywords[1:] else indent
    return ("  " * new_indent + stripped, new_indent)


@click.group()
def yaral():
    """YARA-L specific operations."""
    pass


@yaral.command()
@click.argument("file", type=click.Path(exists=True))
@click.option("--enhanced", is_flag=True, help="Use enhanced parser with full YARA-L 2.0 support")
@click.option("--output", "-o", type=click.Path(), help="Output AST to file")
@click.option(
    "--format",
    type=click.Choice(["json", "yaml", "text"]),
    default="text",
    help="Output format",
)
def parse(file: str, enhanced: bool, output: str | None, format: str):
    """Parse YARA-L file and display AST."""
    try:
        with open(file) as f:
            content = f.read()

        # Choose parser
        if enhanced:
            parser = EnhancedYaraLParser(content)
            click.echo("Using enhanced YARA-L parser...")
        else:
            parser = YaraLParser(content)
            click.echo("Using standard YARA-L parser...")

        # Parse
        ast = parser.parse()

        # Format output
        if format == "json":
            import json

            output_str = json.dumps(ast.__dict__, default=str, indent=2)
        elif format == "yaml":
            import yaml

            output_str = yaml.dump(ast.__dict__, default_flow_style=False)
        else:
            output_str = str(ast)

        # Output
        if output:
            Path(output).write_text(output_str)
            click.echo(f"AST written to {output}")
        else:
            click.echo(output_str)

        click.echo(f"✅ Successfully parsed {len(ast.rules)} rules")

    except Exception as e:
        click.echo(f"❌ Error parsing YARA-L file: {e}", err=True)
        raise click.Abort() from None


@yaral.command()
@click.argument("file", type=click.Path(exists=True))
@click.option("--strict", is_flag=True, help="Treat warnings as errors")
@click.option("--json", "output_json", is_flag=True, help="Output results as JSON")
def validate(file: str, strict: bool, output_json: bool):
    """Validate YARA-L file for semantic correctness."""
    try:
        with open(file) as f:
            content = f.read()

        # Parse
        parser = YaraLParser(content)
        ast = parser.parse()

        # Validate
        validator = YaraLValidator()
        errors, warnings = validator.validate(ast)

        _display_validation_results(file, ast, errors, warnings, strict, output_json)

    except Exception as e:
        click.echo(f"❌ Error validating YARA-L file: {e}", err=True)
        raise click.Abort() from None


@yaral.command()
@click.argument("file", type=click.Path(exists=True))
@click.option("--output", "-o", type=click.Path(), help="Output optimized YARA-L to file")
@click.option("--stats", is_flag=True, help="Show optimization statistics")
@click.option(
    "--dry-run",
    is_flag=True,
    help="Show what would be optimized without making changes",
)
def optimize(file: str, output: str | None, stats: bool, dry_run: bool):
    """Optimize YARA-L rules for better performance."""
    try:
        with open(file) as f:
            content = f.read()

        # Parse
        parser = YaraLParser(content)
        ast = parser.parse()

        # Optimize
        optimizer = YaraLOptimizer()
        optimized_ast, optimization_stats = optimizer.optimize(ast)

        if dry_run:
            click.echo("🔍 Optimization Preview (dry run):")
            click.echo(f"  • Would optimize {optimization_stats.rules_optimized} rules")
            click.echo(f"  • Would simplify {optimization_stats.conditions_simplified} conditions")
            click.echo(f"  • Would optimize {optimization_stats.events_optimized} events")
            click.echo(
                f"  • Would remove {optimization_stats.redundant_checks_removed} redundant checks"
            )
            click.echo(f"  • Would suggest {optimization_stats.indexes_suggested} indexes")
            return

        # Generate optimized code
        generator = YaraLGenerator()
        optimized_code = generator.generate(optimized_ast)

        # Output
        if output:
            Path(output).write_text(optimized_code)
            click.echo(f"✅ Optimized YARA-L written to {output}")
        else:
            click.echo(optimized_code)

        # Show stats
        if stats:
            click.echo("\n📊 Optimization Statistics:")
            click.echo(f"  • Rules optimized: {optimization_stats.rules_optimized}")
            click.echo(f"  • Conditions simplified: {optimization_stats.conditions_simplified}")
            click.echo(f"  • Events optimized: {optimization_stats.events_optimized}")
            click.echo(
                f"  • Redundant checks removed: {optimization_stats.redundant_checks_removed}"
            )
            click.echo(f"  • Indexes suggested: {optimization_stats.indexes_suggested}")
            click.echo(f"  • Time windows optimized: {optimization_stats.time_windows_optimized}")

    except Exception as e:
        click.echo(f"❌ Error optimizing YARA-L file: {e}", err=True)
        raise click.Abort() from None


@yaral.command()
@click.argument("file", type=click.Path(exists=True))
@click.option("--output", "-o", type=click.Path(), help="Output generated YARA-L to file")
@click.option("--format", is_flag=True, help="Format the output code")
def generate(file: str, output: str | None, format: bool):
    """Generate YARA-L code from AST or transform existing rules."""
    try:
        with open(file) as f:
            content = f.read()

        # Parse
        parser = EnhancedYaraLParser(content)
        ast = parser.parse()

        # Generate
        generator = YaraLGenerator()
        code = generator.generate(ast)

        # Format if requested
        if format:
            code = _format_yaral_code(code)

        # Output
        if output:
            Path(output).write_text(code)
            click.echo(f"✅ Generated YARA-L written to {output}")
        else:
            click.echo(code)

        click.echo(f"✅ Successfully generated code for {len(ast.rules)} rules")

    except Exception as e:
        click.echo(f"❌ Error generating YARA-L code: {e}", err=True)
        raise click.Abort() from None


@yaral.command()
@click.argument("file1", type=click.Path(exists=True))
@click.argument("file2", type=click.Path(exists=True))
@click.option("--semantic", is_flag=True, help="Compare semantic meaning, not just syntax")
def compare(file1: str, file2: str, semantic: bool):
    """Compare two YARA-L files for differences."""
    try:
        # Parse both files
        ast1 = _parse_file(file1)
        ast2 = _parse_file(file2)

        # Compare
        if semantic:
            _compare_semantic(ast1, ast2)
        else:
            _compare_structural(ast1, ast2)

    except Exception as e:
        click.echo(f"❌ Error comparing YARA-L files: {e}", err=True)
        raise click.Abort() from None


def _parse_file(file_path: str):
    """Parse a YARA-L file and return its AST."""
    with open(file_path) as f:
        content = f.read()
    parser = YaraLParser(content)
    return parser.parse()


def _compare_semantic(ast1, ast2):
    """Compare two ASTs semantically."""
    generator = YaraLGenerator()
    normalized1 = generator.generate(ast1)
    normalized2 = generator.generate(ast2)

    if normalized1 == normalized2:
        click.echo("✅ Files are semantically equivalent")
    else:
        click.echo("❌ Files are semantically different")


def _compare_structural(ast1, ast2):
    """Compare two ASTs structurally."""
    if len(ast1.rules) != len(ast2.rules):
        click.echo(f"❌ Different number of rules: {len(ast1.rules)} vs {len(ast2.rules)}")
        return

    differences = []
    for _i, (rule1, rule2) in enumerate(zip(ast1.rules, ast2.rules, strict=False)):
        if rule1.name != rule2.name:
            differences.append(f"Rule name: {rule1.name} vs {rule2.name}")

    if differences:
        click.echo("❌ Files have differences:")
        for diff in differences:
            click.echo(f"  • {diff}")
    else:
        click.echo("✅ Files have the same structure")


@yaral.command()
@click.option("--examples", is_flag=True, help="Show example YARA-L rules")
@click.option("--fields", is_flag=True, help="Show valid UDM fields")
@click.option("--functions", is_flag=True, help="Show available aggregation functions")
def info(examples: bool, fields: bool, functions: bool):
    """Show information about YARA-L syntax and features."""
    if examples:
        click.echo("📚 Example YARA-L Rules:\n")
        click.echo(
            """rule suspicious_login_attempts {
    meta:
        author = "security-team"
        description = "Detect multiple failed login attempts"
        severity = "medium"

    events:
        $login.metadata.event_type = "USER_LOGIN"
        $login.security_result.action = "BLOCK"
        $login.principal.user = $user

    match:
        $user over 5m

    condition:
        #login > 5

    outcome:
        $risk_score = 50 + (#login * 10)
        $affected_user = $user
        $login_count = count($login)
}"""
        )

    if fields:
        click.echo("📋 Valid UDM Field Namespaces:\n")
        validator = YaraLValidator()
        for namespace, fields_list in validator.VALID_UDM_FIELDS.items():
            click.echo(f"  {namespace}:")
            for field in fields_list[:5]:  # Show first 5
                click.echo(f"    • {field}")
            if len(fields_list) > 5:
                click.echo(f"    ... and {len(fields_list) - 5} more")

    if functions:
        click.echo("🔧 Available Aggregation Functions:\n")
        functions = [
            ("count", "Count number of events"),
            ("count_distinct", "Count unique values"),
            ("sum", "Sum numeric values"),
            ("avg", "Calculate average"),
            ("min", "Find minimum value"),
            ("max", "Find maximum value"),
            ("array", "Collect values into array"),
            ("array_distinct", "Collect unique values into array"),
            ("string_concat", "Concatenate string values"),
        ]
        for func, desc in functions:
            click.echo(f"  • {func}(): {desc}")

    if not any([examples, fields, functions]):
        click.echo("YARA-L Support Status:")
        click.echo("  ✅ Basic rule parsing")
        click.echo("  ✅ Event patterns and UDM fields")
        click.echo("  ✅ Time windows and correlation")
        click.echo("  ✅ Aggregation functions")
        click.echo("  ✅ Conditional expressions")
        click.echo("  ✅ Semantic validation")
        click.echo("  ✅ Query optimization")
        click.echo("  ✅ Code generation")
        click.echo("\nUse --examples, --fields, or --functions for more details.")



"""Type system for YARA AST semantic validation."""

from yaraast.types.type_system import (
    ArrayType,
    BooleanType,
    DictionaryType,
    DoubleType,
    FunctionType,
    IntegerType,
    ModuleType,
    RangeType,
    StringSetType,
    StringType,
    TypeChecker,
    TypeInference,
    TypeValidator,
    UnknownType,
    YaraType,
)

__all__ = [
    "ArrayType",
    "BooleanType",
    "DictionaryType",
    "DoubleType",
    "FunctionType",
    "IntegerType",
    "ModuleType",
    "RangeType",
    "StringSetType",
    "StringType",
    "TypeChecker",
    "TypeInference",
    "TypeValidator",
    "UnknownType",
    "YaraType",
]



"""YARA-X compatibility checker."""

from __future__ import annotations

import re
from typing import TYPE_CHECKING, Any

from yaraast.ast.expressions import BinaryExpression, Identifier, SetExpression
from yaraast.visitor import ASTVisitor
from yaraast.yarax.feature_flags import YaraXFeatures

if TYPE_CHECKING:
    from yaraast.ast.base import Location, YaraFile
    from yaraast.ast.conditions import OfExpression
    from yaraast.ast.rules import Rule
    from yaraast.ast.strings import HexJump, HexString, PlainString, RegexString


class CompatibilityIssue:
    """Represents a compatibility issue between YARA and YARA-X."""

    def __init__(
        self,
        severity: str,
        location: Location | None,
        issue_type: str,
        message: str,
        suggestion: str = "",
    ) -> None:
        self.severity = severity  # "error", "warning", "info"
        self.location = location
        self.issue_type = issue_type
        self.message = message
        self.suggestion = suggestion

    def __str__(self) -> str:
        loc = f"{self.location.line}:{self.location.column}" if self.location else "unknown"
        return f"[{self.severity.upper()}] {loc}: {self.message}"


class YaraXCompatibilityChecker(ASTVisitor[None]):
    """Check YARA rules for YARA-X compatibility."""

    def __init__(self, features: YaraXFeatures | None = None) -> None:
        self.features = features or YaraXFeatures.yarax_strict()
        self.issues: list[CompatibilityIssue] = []
        self.current_rule: str | None = None

    def check(self, yara_file: YaraFile) -> list[CompatibilityIssue]:
        """Check YARA file for compatibility issues."""
        self.issues.clear()
        self.visit(yara_file)
        return self.issues

    def _add_issue(
        self,
        severity: str,
        issue_type: str,
        message: str,
        suggestion: str = "",
        location: Location | None = None,
    ) -> None:
        """Add a compatibility issue."""
        self.issues.append(
            CompatibilityIssue(
                severity=severity,
                location=location,
                issue_type=issue_type,
                message=message,
                suggestion=suggestion,
            ),
        )

    def visit_rule(self, node: Rule) -> None:
        """Check rule for compatibility issues."""
        self.current_rule = node.name

        # Check for duplicate modifiers
        if self.features.disallow_duplicate_modifiers:
            seen_modifiers = set()
            for modifier in node.modifiers:
                if modifier in seen_modifiers:
                    self._add_issue(
                        "error",
                        "duplicate_modifier",
                        f"Duplicate '{modifier}' modifier in rule '{node.name}'",
                        f"Remove duplicate '{modifier}' modifier",
                        node.location,
                    )
                seen_modifiers.add(modifier)

        # Visit rule components
        # Handle meta as either dict or list
        if isinstance(node.meta, dict):
            # Meta is a dictionary, check values
            for _key, _value in node.meta.items():
                # Check meta values directly
                pass
        elif isinstance(node.meta, list):
            # Meta is a list of Meta objects
            for meta in node.meta:
                if hasattr(meta, "accept"):
                    self.visit(meta)

        for string_def in node.strings:
            self.visit(string_def)

        self.visit(node.condition)

        self.current_rule = None

    def visit_plain_string(self, node: PlainString) -> None:
        """Check plain string compatibility."""
        # Check base64 modifier
        has_base64 = any(m.name in ("base64", "base64wide") for m in node.modifiers)
        if (
            has_base64
            and self.features.minimum_base64_length > 0
            and len(node.value) < self.features.minimum_base64_length
        ):
            self._add_issue(
                "error",
                "base64_too_short",
                f"Base64 pattern '{node.value}' is shorter than minimum length {self.features.minimum_base64_length}",
                f"Use a base64 pattern with at least {self.features.minimum_base64_length} characters",
                node.location,
            )

        # Check XOR with fullword
        has_xor = any(m.name == "xor" for m in node.modifiers)
        has_fullword = any(m.name == "fullword" for m in node.modifiers)

        if (
            has_xor
            and has_fullword
            and self.features.strict_xor_fullword
            and not re.match(r"^[a-zA-Z0-9].*[a-zA-Z0-9]$", node.value)
        ):
            self._add_issue(
                "warning",
                "xor_fullword_boundary",
                f"String '{node.value}' with XOR and fullword may have stricter boundary checking in YARA-X",
                "Ensure string has proper alphanumeric boundaries",
                node.location,
            )

    def visit_regex_string(self, node: RegexString) -> None:
        """Check regex string compatibility."""
        if self.features.strict_regex_escaping:
            self._check_unescaped_braces(node)

        if self.features.validate_escape_sequences:
            self._check_invalid_escape_sequences(node)

    def _check_unescaped_braces(self, node: RegexString) -> None:
        """Check for unescaped { outside of repetition."""
        pattern = node.regex
        i = 0
        while i < len(pattern):
            if pattern[i] == "\\":
                # Skip escaped character
                i += 2
                continue
            if pattern[i] == "{" and not self._is_valid_quantifier(pattern, i):
                self._add_issue(
                    "error",
                    "unescaped_brace",
                    f"Unescaped '{{' in regex pattern at position {i}",
                    "Escape the brace with '\\{'",
                    node.location,
                )
            i += 1

    def _is_valid_quantifier(self, pattern: str, start_pos: int) -> bool:
        """Check if brace at position is part of a valid quantifier."""
        j = start_pos + 1
        if j >= len(pattern):
            return False

        # Skip initial digits
        if not self._skip_digits(pattern, j):
            return False
        j = self._get_position_after_digits(pattern, j)

        if j >= len(pattern):
            return False

        # Check patterns: {n}, {n,}, {n,m}
        if pattern[j] == "}":
            return True  # {n} pattern

        if pattern[j] == "," and j + 1 < len(pattern):
            j += 1
            if pattern[j] == "}":
                return True  # {n,} pattern

            # Check for {n,m} pattern
            if self._skip_digits(pattern, j):
                j = self._get_position_after_digits(pattern, j)
                return j < len(pattern) and pattern[j] == "}"

        return False

    def _skip_digits(self, pattern: str, pos: int) -> bool:
        """Check if position starts with digits."""
        return pos < len(pattern) and pattern[pos].isdigit()

    def _get_position_after_digits(self, pattern: str, start_pos: int) -> int:
        """Get position after consuming all digits."""
        pos = start_pos
        while pos < len(pattern) and pattern[pos].isdigit():
            pos += 1
        return pos

    def _check_invalid_escape_sequences(self, node: RegexString) -> None:
        """Check for invalid escape sequences in regex."""
        invalid_escapes = re.findall(
            r"\\([^\\abfnrtv0-7xdDsSwW.*+?{}()\[\]|^$])",
            node.regex,
        )
        for escape in invalid_escapes:
            self._add_issue(
                "error",
                "invalid_escape",
                f"Invalid escape sequence '\\{escape}' in regex",
                "Remove or fix the escape sequence",
                node.location,
            )

    def visit_hex_string(self, node: HexString) -> None:
        """Check hex string compatibility."""
        for token in node.tokens:
            self.visit(token)

    def visit_hex_jump(self, node: HexJump) -> None:
        """Check hex jump compatibility."""
        if self.features.validate_hex_bounds:
            # YARA-X accepts hex and octal values in bounds
            # This is actually an enhancement, so we might want to note it
            pass

    def visit_of_expression(self, node: OfExpression) -> None:
        """Check 'of' expression compatibility."""
        # YARA-X allows tuples of boolean expressions
        if self.features.allow_tuple_of_expressions and isinstance(
            node.string_set,
            SetExpression,
        ):
            for elem in node.string_set.elements:
                if isinstance(elem, BinaryExpression):
                    self._add_issue(
                        "info",
                        "yarax_feature",
                        "Using YARA-X feature: boolean expressions in 'of' statement",
                        "This feature is not available in original YARA",
                        node.location,
                    )
                    break

        self.visit(node.quantifier)
        self.visit(node.string_set)

    def visit_identifier(self, node: Identifier) -> None:
        """Check for 'with' statement usage."""
        # The 'with' statement would be parsed differently, but we can detect attempts
        if node.name == "with" and self.current_rule and not self.features.allow_with_statement:
            self._add_issue(
                "error",
                "unsupported_feature",
                "'with' statement is only available in YARA-X",
                "Rewrite without using 'with' statement for YARA compatibility",
                node.location,
            )

    def get_report(self) -> dict[str, Any]:
        """Generate compatibility report."""
        errors = [i for i in self.issues if i.severity == "error"]
        warnings = [i for i in self.issues if i.severity == "warning"]
        infos = [i for i in self.issues if i.severity == "info"]

        return {
            "compatible": len(errors) == 0,
            "total_issues": len(self.issues),
            "errors": len(errors),
            "warnings": len(warnings),
            "info": len(infos),
            "issues_by_type": self._group_by_type(),
            "yarax_features_used": self._get_yarax_features_used(),
            "migration_difficulty": self._assess_migration_difficulty(),
        }

    def _group_by_type(self) -> dict[str, list[CompatibilityIssue]]:
        """Group issues by type."""
        grouped: dict[str, list[CompatibilityIssue]] = {}
        for issue in self.issues:
            if issue.issue_type not in grouped:
                grouped[issue.issue_type] = []
            grouped[issue.issue_type].append(issue)
        return grouped

    def _get_yarax_features_used(self) -> list[str]:
        """Get list of YARA-X specific features used."""
        features = []
        for issue in self.issues:
            if issue.issue_type == "yarax_feature":
                features.append(issue.message.split(": ", 1)[1])
        return list(set(features))

    def _assess_migration_difficulty(self) -> str:
        """Assess difficulty of migrating to YARA-X."""
        errors = sum(1 for i in self.issues if i.severity == "error")
        warnings = sum(1 for i in self.issues if i.severity == "warning")

        if errors == 0 and warnings == 0:
            return "trivial"
        if errors == 0:
            return "easy"
        if errors <= 5:
            return "moderate"
        return "difficult"

    # Implement all required abstract methods with default behavior
    def visit_yara_file(self, node) -> None:
        """Visit YARA file."""
        for rule in node.rules:
            self.visit(rule)

    def visit_array_access(self, node) -> None:
        # Not needed for compatibility checking
        pass  # Implementation intentionally empty

    def visit_at_expression(self, node) -> None:
        # Not needed for compatibility checking
        pass  # Implementation intentionally empty

    def visit_binary_expression(self, node) -> None:
        # Not needed for compatibility checking
        pass  # Implementation intentionally empty

    def visit_boolean_literal(self, node) -> None:
        # Not needed for compatibility checking
        pass  # Implementation intentionally empty

    def visit_comment(self, node) -> None:
        # Not needed for compatibility checking
        pass  # Implementation intentionally empty

    def visit_comment_group(self, node) -> None:
        # Not needed for compatibility checking
        pass  # Implementation intentionally empty

    def visit_condition(self, node) -> None:
        # Not needed for compatibility checking
        pass  # Implementation intentionally empty

    def visit_defined_expression(self, node) -> None:
        # Not needed for compatibility checking
        pass  # Implementation intentionally empty

    def visit_dictionary_access(self, node) -> None:
        # Not needed for compatibility checking
        pass  # Implementation intentionally empty

    def visit_double_literal(self, node) -> None:
        # Not needed for compatibility checking
        pass  # Implementation intentionally empty

    def visit_expression(self, node) -> None:
        # Not needed for compatibility checking
        pass  # Implementation intentionally empty

    def visit_extern_import(self, node) -> None:
        # Not needed for compatibility checking
        pass  # Implementation intentionally empty

    def visit_extern_namespace(self, node) -> None:
        # Not needed for compatibility checking
        pass  # Implementation intentionally empty

    def visit_extern_rule(self, node) -> None:
        # Not needed for compatibility checking
        pass  # Implementation intentionally empty

    def visit_extern_rule_reference(self, node) -> None:
        # Not needed for compatibility checking
        pass  # Implementation intentionally empty

    def visit_for_expression(self, node) -> None:
        # Not needed for compatibility checking
        pass  # Implementation intentionally empty

    def visit_for_of_expression(self, node) -> None:
        # Not needed for compatibility checking
        pass  # Implementation intentionally empty

    def visit_function_call(self, node) -> None:
        # Not needed for compatibility checking
        pass  # Implementation intentionally empty

    def visit_hex_alternative(self, node) -> None:
        # Not needed for compatibility checking
        pass  # Implementation intentionally empty

    def visit_hex_byte(self, node) -> None:
        # Not needed for compatibility checking
        pass  # Implementation intentionally empty

    def visit_hex_nibble(self, node) -> None:
        # Not needed for compatibility checking
        pass  # Implementation intentionally empty

    def visit_hex_token(self, node) -> None:
        # Not needed for compatibility checking
        pass  # Implementation intentionally empty

    def visit_hex_wildcard(self, node) -> None:
        # Not needed for compatibility checking
        pass  # Implementation intentionally empty

    def visit_import(self, node) -> None:
        # Not needed for compatibility checking
        pass  # Implementation intentionally empty

    def visit_in_expression(self, node) -> None:
        # Not needed for compatibility checking
        pass  # Implementation intentionally empty

    def visit_in_rule_pragma(self, node) -> None:
        # Not needed for compatibility checking
        pass  # Implementation intentionally empty

    def visit_include(self, node) -> None:
        # Not needed for compatibility checking
        pass  # Implementation intentionally empty

    def visit_integer_literal(self, node) -> None:
        # Not needed for compatibility checking
        pass  # Implementation intentionally empty

    def visit_member_access(self, node) -> None:
        # Not needed for compatibility checking
        pass  # Implementation intentionally empty

    def visit_meta(self, node) -> None:
        # Not needed for compatibility checking
        pass  # Implementation intentionally empty

    def visit_module_reference(self, node) -> None:
        # Not needed for compatibility checking
        pass  # Implementation intentionally empty

    def visit_parentheses_expression(self, node) -> None:
        # Not needed for compatibility checking
        pass  # Implementation intentionally empty

    def visit_pragma(self, node) -> None:
        # Not needed for compatibility checking
        pass  # Implementation intentionally empty

    def visit_pragma_block(self, node) -> None:
        # Not needed for compatibility checking
        pass  # Implementation intentionally empty

    def visit_range_expression(self, node) -> None:
        # Not needed for compatibility checking
        pass  # Implementation intentionally empty

    def visit_regex_literal(self, node) -> None:
        # Not needed for compatibility checking
        pass  # Implementation intentionally empty

    def visit_set_expression(self, node) -> None:
        # Not needed for compatibility checking
        pass  # Implementation intentionally empty

    def visit_string_count(self, node) -> None:
        # Not needed for compatibility checking
        pass  # Implementation intentionally empty

    def visit_string_definition(self, node) -> None:
        # Not needed for compatibility checking
        pass  # Implementation intentionally empty

    def visit_string_identifier(self, node) -> None:
        # Not needed for compatibility checking
        pass  # Implementation intentionally empty

    def visit_string_wildcard(self, node) -> None:
        # Not needed for compatibility checking
        pass  # Implementation intentionally empty

    def visit_string_length(self, node) -> None:
        # Not needed for compatibility checking
        pass  # Implementation intentionally empty

    def visit_string_literal(self, node) -> None:
        # Not needed for compatibility checking
        pass  # Implementation intentionally empty

    def visit_string_modifier(self, node) -> None:
        # Not needed for compatibility checking
        pass  # Implementation intentionally empty

    def visit_string_offset(self, node) -> None:
        # Not needed for compatibility checking
        pass  # Implementation intentionally empty

    def visit_string_operator_expression(self, node) -> None:
        # Not needed for compatibility checking
        pass  # Implementation intentionally empty

    def visit_tag(self, node) -> None:
        # Not needed for compatibility checking
        pass  # Implementation intentionally empty

    def visit_unary_expression(self, node) -> None:
        # Not needed for compatibility checking
        pass  # Implementation intentionally empty


# Alias for compatibility
CompatibilityChecker = YaraXCompatibilityChecker



"""Rule-related AST nodes."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any

from yaraast.ast.base import ASTNode
from yaraast.ast.modifiers import MetaEntry, RuleModifier

if TYPE_CHECKING:
    from yaraast.ast.conditions import Condition
    from yaraast.ast.pragmas import InRulePragma
    from yaraast.ast.strings import StringDefinition


@dataclass
class Import(ASTNode):
    """Import statement node."""

    module: str
    alias: str | None = None  # Support for 'import "module" as alias'

    def accept(self, visitor: Any) -> Any:
        return visitor.visit_import(self)


@dataclass
class Include(ASTNode):
    """Include statement node."""

    path: str

    def accept(self, visitor: Any) -> Any:
        return visitor.visit_include(self)


@dataclass
class Tag(ASTNode):
    """Rule tag node."""

    name: str

    def accept(self, visitor: Any) -> Any:
        return visitor.visit_tag(self)


@dataclass
class Rule(ASTNode):
    """YARA rule node with enhanced modifier and meta support."""

    name: str
    modifiers: list[str | RuleModifier] = field(
        default_factory=list,
    )  # Support both old and new
    tags: list[Tag] = field(default_factory=list)
    meta: dict[str, Any] | list[MetaEntry] = field(
        default_factory=dict,
    )  # Support enhanced meta
    strings: list[StringDefinition] = field(default_factory=list)
    condition: Condition | None = None
    pragmas: list[InRulePragma] = field(default_factory=list)  # In-rule pragmas

    def accept(self, visitor: Any) -> Any:
        return visitor.visit_rule(self)

    @property
    def is_private(self) -> bool:
        """Check if rule is private."""
        for mod in self.modifiers:
            if isinstance(mod, str) and mod == "private":
                return True
            if isinstance(mod, RuleModifier):
                from yaraast.ast.modifiers import RuleModifierType

                if mod.modifier_type == RuleModifierType.PRIVATE:
                    return True
        return False

    @property
    def is_global(self) -> bool:
        """Check if rule is global."""
        for mod in self.modifiers:
            if isinstance(mod, str) and mod == "global":
                return True
            if isinstance(mod, RuleModifier):
                from yaraast.ast.modifiers import RuleModifierType

                if mod.modifier_type == RuleModifierType.GLOBAL:
                    return True
        return False

    def get_meta_entries(self) -> list[MetaEntry]:
        """Get meta entries as enhanced MetaEntry objects."""
        if isinstance(self.meta, list):
            return self.meta
        # Convert dict to MetaEntry list
        return [MetaEntry.from_key_value(k, v) for k, v in self.meta.items()]

    def get_private_meta(self) -> list[MetaEntry]:
        """Get only private meta entries."""
        entries = self.get_meta_entries()
        return [entry for entry in entries if entry.is_private]

    def get_public_meta(self) -> list[MetaEntry]:
        """Get only public meta entries."""
        entries = self.get_meta_entries()
        return [entry for entry in entries if entry.is_public]

    def add_pragma(self, pragma: InRulePragma) -> None:
        """Add a pragma to this rule."""
        self.pragmas.append(pragma)

    def get_pragmas_by_position(self, position: str) -> list[InRulePragma]:
        """Get pragmas by their position in the rule."""
        return [p for p in self.pragmas if p.position == position]



"""Condition-related AST nodes."""

from __future__ import annotations

from dataclasses import dataclass
from typing import Any

from yaraast.ast.expressions import Expression


@dataclass
class Condition(Expression):
    """Base class for conditions."""

    def accept(self, visitor: Any) -> Any:
        return visitor.visit_condition(self)


@dataclass
class ForExpression(Condition):
    """For expression (for any/all i in ...)."""

    quantifier: str
    variable: str
    iterable: Expression
    body: Expression

    def accept(self, visitor: Any) -> Any:
        return visitor.visit_for_expression(self)


@dataclass
class ForOfExpression(Condition):
    """For...of expression (for any of ($a, $b))."""

    quantifier: str
    string_set: Expression
    condition: Expression | None = None

    def accept(self, visitor: Any) -> Any:
        return visitor.visit_for_of_expression(self)


@dataclass
class AtExpression(Condition):
    """At expression ($a at offset)."""

    string_id: str
    offset: Expression

    def accept(self, visitor: Any) -> Any:
        return visitor.visit_at_expression(self)


@dataclass
class InExpression(Condition):
    """In expression ($a in (offset..offset)) or (any of ($a*) in (offset..offset))."""

    string_id: str | None = None
    string_expr: Expression | None = None
    range: Expression | None = None

    def accept(self, visitor: Any) -> Any:
        return visitor.visit_in_expression(self)


@dataclass
class OfExpression(Condition):
    """Of expression (N of ($a, $b, $c))."""

    quantifier: Expression
    string_set: Expression

    def accept(self, visitor: Any) -> Any:
        return visitor.visit_of_expression(self)



"""Error-tolerant YARA lexer that continues parsing despite errors."""

from dataclasses import dataclass

from yaraast.lexer.lexer import Lexer, LexerError
from yaraast.lexer.tokens import Token, TokenType


@dataclass
class LexerErrorInfo:
    """Detailed information about a lexer error."""

    message: str
    line: int
    column: int
    context: str
    suggestion: str | None = None
    severity: str = "error"  # error, warning, info

    def format_error(self) -> str:
        """Format error for display."""
        lines = []
        lines.append(f"\n{'=' * 60}")
        lines.append(f"❌ {self.severity.upper()}: {self.message}")
        lines.append(f"📍 Location: Line {self.line}, Column {self.column}")

        if self.context:
            lines.append("\n📄 Context:")
            # Show the problematic line with markers
            context_lines = self.context.split("\n")
            for i, line in enumerate(context_lines):
                if i == 1:  # Middle line is the error line
                    lines.append(f"    {self.line:4d} | {line}")
                    # Add pointer to the error position
                    pointer = " " * (10 + self.column - 1) + "^" + "~" * 5
                    lines.append(f"         | {pointer}")
                else:
                    line_num = self.line - 1 + i
                    lines.append(f"    {line_num:4d} | {line}")

        if self.suggestion:
            lines.append(f"\n💡 Suggestion: {self.suggestion}")

        lines.append("=" * 60)
        return "\n".join(lines)


class ErrorTolerantLexer(Lexer):
    """Lexer that collects errors and continues parsing."""

    def __init__(self, text: str, max_errors: int = 100) -> None:
        super().__init__(text)
        self.errors: list[LexerErrorInfo] = []
        self.max_errors = max_errors
        self.original_text = text

    def tokenize(self) -> tuple[list[Token], list[LexerErrorInfo]]:
        """Tokenize text and return both tokens and errors."""
        self.tokens = []  # Use self.tokens so _is_hex_string_context() works
        self.position = 0
        self.line = 1
        self.column = 1
        self.errors = []

        while self.position < len(self.text):
            if len(self.errors) >= self.max_errors:
                self._add_error(
                    f"Too many errors (>{self.max_errors}), stopping",
                    severity="error",
                )
                break

            # Skip whitespace and comments first
            self._skip_whitespace_and_comments()

            if self.position >= len(self.text):
                break

            try:
                token = self._next_token()
                if token:
                    self.tokens.append(token)
            except LexerError as e:
                # Record the error with context
                self._add_error(str(e).split(": ", 1)[1] if ": " in str(e) else str(e))
                # Try to recover
                self._recover_from_error()

        # Add EOF token
        self.tokens.append(Token(TokenType.EOF, None, self.line, self.column))
        return self.tokens, self.errors

    def _add_error(
        self,
        message: str,
        severity: str = "error",
        suggestion: str | None = None,
    ) -> None:
        """Add an error with context."""
        # Get context lines
        lines = self.original_text.split("\n")
        context_lines = []

        # Get line before, current line, and line after
        for i in range(max(0, self.line - 2), min(len(lines), self.line + 1)):
            if i < len(lines):
                context_lines.append(lines[i])

        context = "\n".join(context_lines)

        error = LexerErrorInfo(
            message=message,
            line=self.line,
            column=self.column,
            context=context,
            suggestion=suggestion,
            severity=severity,
        )
        self.errors.append(error)

    def _recover_from_error(self) -> None:
        """Try to recover from an error and continue parsing."""
        char = self._current_char()

        # Special recovery for common cases
        if char == '"':
            # Unterminated string - skip to next line or next quote
            self._recover_from_unterminated_string()
        elif char == "/":
            # Possible regex error - skip to next space
            self._skip_to_whitespace()
        elif char == "{":
            # Hex string error - skip to closing brace
            self._skip_to_char("}")
            self._advance()
        else:
            # Default: skip current character and continue
            self._advance()

    def _recover_from_unterminated_string(self) -> None:
        """Recover from unterminated string error."""
        # Skip to end of line or next string delimiter
        while self._current_char() and self._current_char() not in '\n"':
            self._advance()

        if self._current_char() in ('"', "\n"):
            self._advance()  # Skip the quote or newline

    def _skip_to_whitespace(self) -> None:
        """Skip to next whitespace character."""
        while self._current_char() and not self._current_char().isspace():
            self._advance()

    def _skip_to_char(self, target: str) -> None:
        """Skip to target character."""
        while self._current_char() and self._current_char() != target:
            self._advance()

    def _read_string(self) -> Token:
        """Override to handle problematic strings better."""
        start_line = self.line
        start_column = self.column
        value = ""

        self._advance()  # skip opening quote

        while self._current_char() and self._current_char() != '"':
            if self._current_char() == "\\":
                # Look ahead to see what follows the backslash
                self._advance()
                next_char = self._current_char()

                if next_char == "\\":
                    value += "\\"
                elif next_char == '"':
                    # Check for special case: \" at end of string like "\TEMP\"
                    look_ahead_pos = self.position + 1
                    if look_ahead_pos < len(self.text):
                        chars_after = self.text[look_ahead_pos : look_ahead_pos + 20]
                        # Skip whitespace
                        i = 0
                        while i < len(chars_after) and chars_after[i] in " \t":
                            i += 1
                        chars_after = chars_after[i:]

                        # Check if this looks like end of string
                        if not chars_after or chars_after.startswith(
                            (
                                "ascii",
                                "wide",
                                "nocase",
                                "fullword",
                                "xor",
                                "base64",
                                "\n",
                                "\r",
                            ),
                        ):
                            # Treat as Windows path with backslash at end
                            value += "\\"
                            self._advance()  # consume quote

                            # Add warning about non-standard syntax
                            self._add_error(
                                'String ends with \\" which is non-standard YARA syntax',
                                severity="warning",
                                suggestion='Consider using "\\\\" for literal backslash or escaping properly',
                            )
                            break
                        # Normal escaped quote
                        value += '"'
                    else:
                        value += "\\"
                        self._advance()
                        break
                elif next_char == "n":
                    value += "\n"
                elif next_char == "r":
                    value += "\r"
                elif next_char == "t":
                    value += "\t"
                elif next_char == "x" and self.position + 2 < len(self.text):
                    # Hex escape
                    hex_digits = self.text[self.position + 1 : self.position + 3]
                    if all(c in "0123456789abcdefABCDEF" for c in hex_digits):
                        value += chr(int(hex_digits, 16))
                        self._advance()  # skip first hex digit
                        self._advance()  # skip second hex digit
                    else:
                        value += "\\" + next_char
                elif next_char is None:
                    value += "\\"
                    break
                else:
                    # Unknown escape - keep literal
                    value += "\\" + next_char
            else:
                value += self._current_char()

            self._advance()

        if not self._current_char():
            # Unterminated string
            self._add_error(
                "Unterminated string",
                suggestion='Add closing quote " to terminate the string',
            )
            # Return what we have
            return Token(TokenType.STRING, value, start_line, start_column)

        self._advance()  # skip closing quote
        return Token(TokenType.STRING, value, start_line, start_column)



"""Formatting provider for YARA Language Server."""

from __future__ import annotations

from typing import TYPE_CHECKING

from lsprotocol.types import Position, Range, TextEdit

from yaraast.codegen.generator import CodeGenerator
from yaraast.parser.parser import Parser

if TYPE_CHECKING:
    pass


class FormattingProvider:
    """Provides code formatting functionality."""

    def format_document(self, text: str) -> list[TextEdit]:
        """
        Format the entire document.

        Args:
            text: The YARA source code

        Returns:
            List of text edits to apply
        """
        try:
            # Parse the document
            parser = Parser(text)
            ast = parser.parse()

            # Generate formatted code
            generator = CodeGenerator()
            formatted_text = generator.generate(ast)

            # Calculate the range for the entire document
            lines = text.split("\n")
            doc_range = Range(
                start=Position(line=0, character=0),
                end=Position(line=len(lines), character=0),
            )

            # Return a single edit that replaces the entire document
            return [TextEdit(range=doc_range, new_text=formatted_text)]

        except Exception:
            # If formatting fails, return no edits
            return []

    def format_range(
        self,
        text: str,
        start: Position,
        end: Position,
    ) -> list[TextEdit]:
        """
        Format a specific range in the document.

        Args:
            text: The YARA source code
            start: Start position of the range
            end: End position of the range

        Returns:
            List of text edits to apply
        """
        # For now, format the entire document
        # A more sophisticated implementation would only format the selected range
        return self.format_document(text)



"""YARA dialect support for YARA, YARA-X, and YARA-L."""

from enum import Enum, auto


class YaraDialect(Enum):
    """YARA dialect enumeration."""

    YARA = auto()  # Standard YARA
    YARA_X = auto()  # YARA-X (VirusTotal's next-gen YARA)
    YARA_L = auto()  # YARA-L (Google Chronicle)


def detect_dialect(content: str) -> YaraDialect:
    """Detect the YARA dialect from content.

    Args:
        content: The rule content to analyze

    Returns:
        The detected YARA dialect

    """
    # Quick heuristics for dialect detection
    content_lower = content.lower()

    # YARA-L specific keywords
    yaral_keywords = [
        "events:",
        "match:",
        "outcome:",
        "options:",
        "metadata.event_type",
        "principal.",
        "target.",
        "udm.",
        "over ",
        "nocase",
        "count_distinct",
        "array_distinct",
        "earliest",
        "latest",
    ]

    # YARA-X specific features (if any distinct ones exist)
    yarax_keywords = [
        # Add YARA-X specific features when identified
    ]

    # Check for YARA-L indicators
    for keyword in yaral_keywords:
        if keyword in content_lower:
            return YaraDialect.YARA_L

    # Check for YARA-X indicators
    for keyword in yarax_keywords:
        if keyword in content_lower:
            return YaraDialect.YARA_X

    # Default to standard YARA
    return YaraDialect.YARA


__all__ = ["YaraDialect", "detect_dialect"]



"""YARA-X compatibility module."""

from yaraast.yarax.compatibility_checker import YaraXCompatibilityChecker
from yaraast.yarax.feature_flags import YaraXFeatures
from yaraast.yarax.syntax_adapter import YaraXSyntaxAdapter

__all__ = ["YaraXCompatibilityChecker", "YaraXFeatures", "YaraXSyntaxAdapter"]



"""Simple AST nodes without dataclass complexity."""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any


class ASTNode(ABC):
    """Base class for all AST nodes."""

    def __init__(self, **kwargs) -> None:
        self.location = kwargs.get("location")

    @abstractmethod
    def accept(self, visitor: Any) -> Any:
        """Accept a visitor for the visitor pattern."""


class Expression(ASTNode):
    """Base class for expressions."""

    def accept(self, visitor: Any) -> Any:
        return visitor.visit_expression(self)


class StringIdentifier(Expression):
    """String identifier like $a."""

    def __init__(self, name: str, **kwargs) -> None:
        super().__init__(**kwargs)
        self.name = name

    def accept(self, visitor: Any) -> Any:
        return visitor.visit_string_identifier(self)


class BooleanLiteral(Expression):
    """Boolean literal."""

    def __init__(self, value: bool, **kwargs) -> None:
        super().__init__(**kwargs)
        self.value = value

    def accept(self, visitor: Any) -> Any:
        return visitor.visit_boolean_literal(self)


class IntegerLiteral(Expression):
    """Integer literal."""

    def __init__(self, value: int, **kwargs) -> None:
        super().__init__(**kwargs)
        self.value = value

    def accept(self, visitor: Any) -> Any:
        return visitor.visit_integer_literal(self)


class Identifier(Expression):
    """Generic identifier."""

    def __init__(self, name: str, **kwargs) -> None:
        super().__init__(**kwargs)
        self.name = name

    def accept(self, visitor: Any) -> Any:
        return visitor.visit_identifier(self)


class BinaryExpression(Expression):
    """Binary expression."""

    def __init__(self, left: Expression, operator: str, right: Expression, **kwargs) -> None:
        super().__init__(**kwargs)
        self.left = left
        self.operator = operator
        self.right = right

    def accept(self, visitor: Any) -> Any:
        return visitor.visit_binary_expression(self)


class Tag(ASTNode):
    """Rule tag."""

    def __init__(self, name: str, **kwargs) -> None:
        super().__init__(**kwargs)
        self.name = name

    def accept(self, visitor: Any) -> Any:
        return visitor.visit_tag(self)


class StringModifier(ASTNode):
    """String modifier."""

    def __init__(self, name: str, value: Any = None, **kwargs) -> None:
        super().__init__(**kwargs)
        self.name = name
        self.value = value

    def accept(self, visitor: Any) -> Any:
        return visitor.visit_string_modifier(self)


class StringDefinition(ASTNode):
    """Base string definition."""

    def __init__(
        self,
        identifier: str,
        modifiers: list[StringModifier] | None = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.identifier = identifier
        self.modifiers = modifiers or []

    def accept(self, visitor: Any) -> Any:
        return visitor.visit_string_definition(self)


class PlainString(StringDefinition):
    """Plain string."""

    def __init__(
        self,
        identifier: str,
        value: str,
        modifiers: list[StringModifier] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(identifier, modifiers, **kwargs)
        self.value = value

    def accept(self, visitor: Any) -> Any:
        return visitor.visit_plain_string(self)


class HexToken(ASTNode):
    """Base hex token."""

    def accept(self, visitor: Any) -> Any:
        return visitor.visit_hex_token(self)


class HexByte(HexToken):
    """Hex byte."""

    def __init__(self, value: int, **kwargs) -> None:
        super().__init__(**kwargs)
        self.value = value

    def accept(self, visitor: Any) -> Any:
        return visitor.visit_hex_byte(self)


class HexWildcard(HexToken):
    """Hex wildcard."""

    def accept(self, visitor: Any) -> Any:
        return visitor.visit_hex_wildcard(self)


class HexString(StringDefinition):
    """Hex string."""

    def __init__(
        self,
        identifier: str,
        tokens: list[HexToken],
        modifiers: list[StringModifier] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(identifier, modifiers, **kwargs)
        self.tokens = tokens

    def accept(self, visitor: Any) -> Any:
        return visitor.visit_hex_string(self)


class RegexString(StringDefinition):
    """Regex string."""

    def __init__(
        self,
        identifier: str,
        regex: str,
        modifiers: list[StringModifier] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(identifier, modifiers, **kwargs)
        self.regex = regex

    def accept(self, visitor: Any) -> Any:
        return visitor.visit_regex_string(self)


class Rule(ASTNode):
    """YARA rule."""

    def __init__(
        self,
        name: str,
        modifiers: list[str] | None = None,
        tags: list[Tag] | None = None,
        meta: dict[str, Any] | None = None,
        strings: list[StringDefinition] | None = None,
        condition: Expression | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.name = name
        self.modifiers = modifiers or []
        self.tags = tags or []
        self.meta = meta or {}
        self.strings = strings or []
        self.condition = condition

    def accept(self, visitor: Any) -> Any:
        return visitor.visit_rule(self)



"""Definition provider for YARA Language Server."""

from __future__ import annotations

from typing import TYPE_CHECKING

from lsprotocol.types import Location, Position, Range

from yaraast.lsp.utils import get_word_at_position
from yaraast.parser.parser import Parser

if TYPE_CHECKING:
    pass


class DefinitionProvider:
    """Provides go-to-definition functionality."""

    def get_definition(
        self,
        text: str,
        position: Position,
        uri: str,
    ) -> Location | list[Location] | None:
        """
        Get definition location for the symbol at the given position.

        Args:
            text: The YARA source code
            position: The cursor position
            uri: The document URI

        Returns:
            Location of the definition or None
        """
        word, _ = get_word_at_position(text, position)

        if not word:
            return None

        # Handle string identifiers
        if (
            word.startswith("$")
            or word.startswith("#")
            or word.startswith("@")
            or word.startswith("!")
        ):
            # Normalize to base identifier
            base_identifier = word.lstrip("#@!")
            if not base_identifier.startswith("$"):
                base_identifier = f"${base_identifier}"

            return self._find_string_definition(text, base_identifier, uri)

        # Handle rule references
        return self._find_rule_definition(text, word, uri)

    def _find_string_definition(
        self,
        text: str,
        identifier: str,
        uri: str,
    ) -> Location | None:
        """Find the definition of a string identifier."""
        try:
            parser = Parser(text)
            ast = parser.parse()

            lines = text.split("\n")
            for rule in ast.rules:
                for string_def in rule.strings:
                    if string_def.identifier == identifier:
                        # Find the line where this string is defined
                        for line_num, line in enumerate(lines):
                            if identifier in line and "=" in line:
                                col = line.index(identifier)
                                return Location(
                                    uri=uri,
                                    range=Range(
                                        start=Position(line=line_num, character=col),
                                        end=Position(
                                            line=line_num,
                                            character=col + len(identifier),
                                        ),
                                    ),
                                )

        except Exception:
            pass

        return None

    def _find_rule_definition(self, text: str, rule_name: str, uri: str) -> Location | None:
        """Find the definition of a rule."""
        try:
            parser = Parser(text)
            ast = parser.parse()

            lines = text.split("\n")
            for rule in ast.rules:
                if rule.name == rule_name:
                    # Find the line where this rule is defined
                    for line_num, line in enumerate(lines):
                        if "rule" in line and rule_name in line:
                            col = line.index(rule_name)
                            return Location(
                                uri=uri,
                                range=Range(
                                    start=Position(line=line_num, character=col),
                                    end=Position(line=line_num, character=col + len(rule_name)),
                                ),
                            )

        except Exception:
            pass

        return None



"""LSP server command for CLI."""

import click
from rich.console import Console

console = Console()


@click.command()
@click.option(
    "--stdio",
    is_flag=True,
    default=True,
    help="Use stdio for communication (default)",
)
@click.option("--tcp", type=int, help="Use TCP on specified port")
@click.option("--host", default="127.0.0.1", help="Host for TCP mode")
def lsp(stdio: bool, tcp: int | None, host: str) -> None:
    """Start the YARA Language Server.

    The language server provides IDE features like autocomplete, diagnostics,
    hover information, go-to-definition, and more for YARA files.

    Examples:
        yaraast lsp --stdio           # Start with stdio (default)
        yaraast lsp --tcp 5007        # Start on TCP port 5007
    """
    try:
        from yaraast.lsp.server import create_server

        console.print("[green]🚀 Starting YARAAST Language Server...[/green]")

        server = create_server()

        if tcp:
            console.print(f"[blue]📡 Listening on {host}:{tcp}[/blue]")
            server.start_tcp(host, tcp)
        else:
            console.print("[blue]📡 Using stdio for communication[/blue]")
            server.start_io()

    except ImportError as e:
        console.print(f"[red]❌ Missing dependency: {e}[/red]")
        console.print("\nInstall LSP dependencies with:")
        console.print("  pip install 'yaraast[lsp]'")
        console.print("\nOr install pygls manually:")
        console.print("  pip install pygls lsprotocol")
        raise click.Abort from None

    except Exception as e:
        console.print(f"[red]❌ Error starting LSP server: {e}[/red]")
        raise click.Abort from None



"""Better parser implementation that merges old logic."""

from __future__ import annotations

from yaraast.ast.base import *
from yaraast.ast.conditions import *
from yaraast.ast.expressions import *
from yaraast.ast.meta import *
from yaraast.ast.modules import *
from yaraast.ast.operators import *
from yaraast.ast.rules import *
from yaraast.ast.strings import *
from yaraast.lexer import Lexer, Token, TokenType


class Parser:
    """Better YARA parser implementation."""

    def __init__(self) -> None:
        self.tokens: list[Token] = []
        self.position = 0
        self.text = ""

    def parse(self, text: str) -> YaraFile:
        """Parse YARA text and return AST."""
        lexer = Lexer(text)
        self.tokens = lexer.tokenize()
        self.position = 0
        self.text = text

        return self._parse_yara_file()

    def _parse_yara_file(self) -> YaraFile:
        """Parse complete YARA file."""
        imports = []
        includes = []
        rules = []

        while not self._is_at_end():
            token = self._current_token()
            if not token:
                break

            if token.type == TokenType.IMPORT:
                self._advance()
                imports.append(self._parse_import())
            elif token.type == TokenType.INCLUDE:
                self._advance()
                includes.append(self._parse_include())
            elif token.type in (TokenType.RULE, TokenType.PRIVATE, TokenType.GLOBAL):
                rules.append(self._parse_rule())
            else:
                self._advance()  # Skip unknown tokens

        return YaraFile(imports=imports, includes=includes, rules=rules)

    def _parse_rule(self) -> Rule:
        """Parse a rule."""
        modifiers = []

        # Parse modifiers
        while self._current_token() and self._current_token().type in (
            TokenType.PRIVATE,
            TokenType.GLOBAL,
        ):
            modifiers.append(self._current_token().value)
            self._advance()

        # Expect 'rule'
        if not self._match(TokenType.RULE):
            msg = "Expected 'rule'"
            raise Exception(msg)

        # Parse rule name
        if not self._current_token() or self._current_token().type != TokenType.IDENTIFIER:
            msg = "Expected rule name"
            raise Exception(msg)

        name = self._current_token().value
        self._advance()

        # Parse tags
        tags = []
        if self._match(TokenType.COLON):
            while self._current_token() and self._current_token().type == TokenType.IDENTIFIER:
                from yaraast.ast.rules import Tag

                tags.append(Tag(name=self._current_token().value))
                self._advance()

        # Expect '{'
        if not self._match(TokenType.LBRACE):
            msg = "Expected '{'"
            raise Exception(msg)

        # Parse sections
        meta = {}
        strings = []
        condition = None

        while not self._check(TokenType.RBRACE) and not self._is_at_end():
            if self._match(TokenType.META):
                if not self._match(TokenType.COLON):
                    msg = "Expected ':' after 'meta'"
                    raise Exception(msg)
                meta_list = self._parse_meta_section()
                # Convert list of Meta objects to dict
                for m in meta_list:
                    meta[m.key] = m.value
            elif self._match(TokenType.STRINGS):
                if not self._match(TokenType.COLON):
                    msg = "Expected ':' after 'strings'"
                    raise Exception(msg)
                strings = self._parse_strings_section()
            elif self._match(TokenType.CONDITION):
                if not self._match(TokenType.COLON):
                    msg = "Expected ':' after 'condition'"
                    raise Exception(msg)
                condition = self._parse_condition()
            else:
                # Try to skip unrecognized tokens until we find something we know or '}'
                if self._current_token() and self._current_token().type != TokenType.RBRACE:
                    self._advance()
                    continue
                break  # Stop if we find '}' or reach end

        # Expect '}'
        if not self._match(TokenType.RBRACE):
            # Try to recover by finding the closing brace
            current = self._current_token()
            attempts = 0
            while not self._is_at_end() and attempts < 1000:
                if self._match(TokenType.RBRACE):
                    break
                self._advance()
                attempts += 1
            else:
                msg = f"Expected '}}' but found {current.type if current else 'EOF'} at position {self.position}"
                raise Exception(
                    msg,
                )

        # Condition is required
        if condition is None:
            condition = BooleanLiteral(value=True)  # Default to true

        return Rule(
            name=name,
            modifiers=modifiers,
            tags=tags,
            meta=meta,
            strings=strings,
            condition=condition,
        )

    def _parse_condition(self) -> Expression:
        """Parse condition expression."""
        # Check if condition body is empty (next token is closing brace)
        if self._check(TokenType.RBRACE):
            # Return default true condition for empty condition bodies
            return BooleanLiteral(value=True)
        return self._parse_or_expression()

    def _parse_or_expression(self) -> Expression:
        """Parse OR expression."""
        expr = self._parse_and_expression()

        while self._match(TokenType.OR):
            right = self._parse_and_expression()
            expr = BinaryExpression(left=expr, operator="or", right=right)

        # Check for range operator (..)
        if self._match(TokenType.DOUBLE_DOT):
            high = self._parse_and_expression()
            from yaraast.ast.expressions import RangeExpression

            return RangeExpression(low=expr, high=high)

        return expr

    def _parse_and_expression(self) -> Expression:
        """Parse AND expression."""
        expr = self._parse_bitwise_expression()

        while self._match(TokenType.AND):
            right = self._parse_bitwise_expression()
            expr = BinaryExpression(left=expr, operator="and", right=right)

        return expr

    def _parse_bitwise_expression(self) -> Expression:
        """Parse bitwise expressions (&, |, ^, <<, >>)."""
        expr = self._parse_not_expression()

        while self._match(
            TokenType.BITWISE_AND,
            TokenType.BITWISE_OR,
            TokenType.XOR,
            TokenType.SHIFT_LEFT,
            TokenType.SHIFT_RIGHT,
        ):
            operator = self._previous().value
            right = self._parse_not_expression()
            expr = BinaryExpression(left=expr, operator=operator, right=right)

        return expr

    def _parse_not_expression(self) -> Expression:
        """Parse NOT expression."""
        if self._match(TokenType.NOT):
            operand = self._parse_not_expression()
            return UnaryExpression(operator="not", operand=operand)

        return self._parse_relational_expression()

    def _parse_additive_expression(self) -> Expression:
        """Parse additive expressions (+ and -)."""
        expr = self._parse_multiplicative_expression()

        while self._match(TokenType.PLUS, TokenType.MINUS):
            operator = self._previous().value
            right = self._parse_multiplicative_expression()
            expr = BinaryExpression(left=expr, operator=operator, right=right)

        return expr

    def _parse_multiplicative_expression(self) -> Expression:
        """Parse multiplicative expressions (*, /, %)."""
        expr = self._parse_primary_expression()

        while self._match(TokenType.MULTIPLY, TokenType.DIVIDE, TokenType.MODULO):
            operator = self._previous().value
            right = self._parse_primary_expression()
            expr = BinaryExpression(left=expr, operator=operator, right=right)

        return expr

    def _parse_relational_expression(self) -> Expression:
        """Parse relational expressions."""
        expr = self._parse_additive_expression()

        while True:
            if self._match(
                TokenType.LT,
                TokenType.LE,
                TokenType.GT,
                TokenType.GE,
                TokenType.EQ,
                TokenType.NEQ,
                TokenType.CONTAINS,
                TokenType.MATCHES,
                TokenType.STARTSWITH,
                TokenType.ENDSWITH,
                TokenType.ICONTAINS,
                TokenType.ISTARTSWITH,
                TokenType.IENDSWITH,
                TokenType.IEQUALS,
            ):
                operator = self._previous().value.lower()
                right = self._parse_additive_expression()
                expr = BinaryExpression(left=expr, operator=operator, right=right)
            elif self._match(TokenType.AT):
                # Handle 'at' operator (e.g., $string at offset)
                if isinstance(expr, StringIdentifier):
                    # Parse the offset expression (could be complex like filesize - 42)
                    offset = self._parse_additive_expression()
                    from yaraast.ast.conditions import AtExpression

                    expr = AtExpression(string_id=expr.name, offset=offset)
                else:
                    msg = "'at' operator requires string identifier on left side"
                    raise ValueError(
                        msg,
                    )
            elif self._match(TokenType.IN):
                # Handle 'in' operator (e.g., $string in (0..100))
                if isinstance(expr, StringIdentifier):
                    range_expr = self._parse_additive_expression()
                    from yaraast.ast.conditions import InExpression

                    expr = InExpression(string_id=expr.name, range=range_expr)
                else:
                    # Could be other 'in' usage, treat as binary operator
                    right = self._parse_additive_expression()
                    expr = BinaryExpression(left=expr, operator="in", right=right)
            else:
                break

        return expr

    def _parse_primary_expression(self) -> Expression:
        """Parse primary expression."""
        # Handle 'defined' operator
        if self._match(TokenType.DEFINED):
            # Parse the full expression that follows 'defined' (could be complex like pe.sections[0].name)
            expr = self._parse_postfix_expression()
            return DefinedExpression(expression=expr)

        # String identifiers
        if self._match(TokenType.STRING_IDENTIFIER):
            return StringIdentifier(name=self._previous().value)

        # String count (#string)
        if self._match(TokenType.STRING_COUNT):
            string_id = self._previous().value  # e.g., "#a"
            # Remove the # prefix to get just the identifier
            string_id = string_id.removeprefix("#")
            return StringCount(string_id=string_id)

        # String offset (@string or @string[index])
        if self._match(TokenType.STRING_OFFSET):
            string_id = self._previous().value  # e.g., "@a"
            # Remove the @ prefix to get just the identifier
            string_id = string_id.removeprefix("@")
            index = None
            if self._match(TokenType.LBRACKET):
                index = self._parse_expression()
                if not self._match(TokenType.RBRACKET):
                    msg = "Expected ']' after string offset index"
                    raise Exception(msg)
            return StringOffset(string_id=string_id, index=index)

        # String length (!string or !string[index])
        if self._match(TokenType.STRING_LENGTH):
            string_id = self._previous().value  # e.g., "!a"
            # Remove the ! prefix to get just the identifier
            string_id = string_id.removeprefix("!")
            index = None
            if self._match(TokenType.LBRACKET):
                index = self._parse_expression()
                if not self._match(TokenType.RBRACKET):
                    msg = "Expected ']' after string length index"
                    raise Exception(msg)
            return StringLength(string_id=string_id, index=index)

        # Skip - integers are handled later to check for "N of" pattern

        if self._match(TokenType.DOUBLE):
            return DoubleLiteral(value=self._previous().value)

        if self._match(TokenType.STRING):
            return StringLiteral(value=self._previous().value)

        if self._match(TokenType.BOOLEAN_TRUE):
            return BooleanLiteral(value=True)

        if self._match(TokenType.BOOLEAN_FALSE):
            return BooleanLiteral(value=False)

        # Regex literals
        if self._match(TokenType.REGEX):
            # Parse regex: pattern with optional modifiers
            regex_val = self._previous().value
            pattern = regex_val
            modifiers = ""

            # The lexer returns pattern or pattern\x00modifiers
            if "\x00" in regex_val:
                parts = regex_val.split("\x00", 1)
                pattern = parts[0]
                modifiers = parts[1] if len(parts) > 1 else ""

            return RegexLiteral(pattern=pattern, modifiers=modifiers)

        # Identifiers and keywords that can be used as identifiers
        if self._match(TokenType.IDENTIFIER, TokenType.FILESIZE, TokenType.ENTRYPOINT):
            return self._parse_postfix_from_identifier(self._previous().value)

        # Handle 'them' as special identifier
        if self._match(TokenType.THEM):
            return Identifier(name="them")

        # Parentheses
        if self._match(TokenType.LPAREN):
            expr = self._parse_or_expression()

            # Handle range with dash syntax (0-100) as alternative to (0..100)
            if self._current_token() and self._current_token().type == TokenType.MINUS:
                self._advance()
                high = self._parse_or_expression()
                from yaraast.ast.expressions import RangeExpression

                expr = RangeExpression(low=expr, high=high)

            if not self._match(TokenType.RPAREN):
                current = self._current_token()
                # Try to recover by skipping unexpected tokens
                while not self._is_at_end() and not self._check(TokenType.RPAREN):
                    self._advance()
                    if self._match(TokenType.RPAREN):
                        break
                else:
                    msg = f"Expected ')' in parentheses at position {self.position}, got {current.type if current else 'EOF'}"
                    raise Exception(
                        msg,
                    )
            return ParenthesesExpression(expression=expr)

        # Unary minus for negative numbers
        if self._match(TokenType.MINUS):
            if self._match(TokenType.INTEGER):
                return IntegerLiteral(value=-self._previous().value)
            if self._match(TokenType.DOUBLE):
                return DoubleLiteral(value=-self._previous().value)
            msg = "Expected number after '-'"
            raise Exception(msg)

        # Special expressions: for
        if self._match(TokenType.FOR):
            return self._parse_for_expression()

        # Special expressions: quantifier of (any of, all of, 2 of, etc.)
        if self._match(TokenType.ANY, TokenType.ALL):
            quantifier = self._previous().value
            if self._match(TokenType.OF):
                return self._parse_of_expression(quantifier)
            # Put the token back, it's just an identifier
            self.position -= 1
            return Identifier(name=quantifier)

        # Check for integer - might be part of "N of" expression
        if self._match(TokenType.INTEGER):
            value = self._previous().value
            # Look ahead to see if "of" follows
            if self._match(TokenType.OF):
                return self._parse_of_expression(value)
            # Just a regular integer literal
            return IntegerLiteral(value=value)

        msg = f"Unexpected token in expression: {self._current_token()}"
        raise Exception(msg)

    def _parse_postfix_expression(self) -> Expression:
        """Parse postfix expression starting from primary."""
        # Get base identifier (could be regular identifier or string identifier)
        if self._match(TokenType.IDENTIFIER):
            return self._parse_postfix_from_identifier(self._previous().value)
        if self._match(TokenType.STRING_IDENTIFIER):
            # For string identifiers like $string1, return a StringIdentifier directly
            return StringIdentifier(name=self._previous().value)
        msg = "Expected identifier in postfix expression"
        raise ValueError(msg)

    def _parse_postfix_from_identifier(self, name: str) -> Expression:
        """Parse postfix operations from an identifier."""
        expr = Identifier(name=name)

        while True:
            if self._match(TokenType.DOT):
                if not self._match(TokenType.IDENTIFIER):
                    msg = "Expected member name after '.'"
                    raise Exception(msg)
                member = self._previous().value
                expr = MemberAccess(object=expr, member=member)
            elif self._match(TokenType.LBRACKET):
                index = self._parse_primary_expression()
                if not self._match(TokenType.RBRACKET):
                    msg = "Expected ']'"
                    raise Exception(msg)
                if isinstance(index, StringLiteral):
                    expr = DictionaryAccess(object=expr, key=index.value)
                else:
                    expr = ArrayAccess(array=expr, index=index)
            elif self._match(TokenType.LPAREN):
                # Function call
                args = []
                while not self._check(TokenType.RPAREN) and not self._is_at_end():
                    # Try to parse argument, but be tolerant of errors
                    try:
                        args.append(self._parse_or_expression())
                    except (ValueError, TypeError, AttributeError):
                        # If parsing fails, try to recover by finding comma or closing paren
                        while not self._is_at_end():
                            if self._check(TokenType.COMMA) or self._check(
                                TokenType.RPAREN,
                            ):
                                break
                            self._advance()

                    if not self._match(TokenType.COMMA):
                        break

                if not self._match(TokenType.RPAREN):
                    # Try to recover by finding the closing parenthesis
                    current = self._current_token()
                    attempts = 0
                    while not self._is_at_end() and attempts < 100:
                        if self._match(TokenType.RPAREN):
                            break
                        self._advance()
                        attempts += 1
                    else:
                        msg = f"Expected ')' in function call at position {self.position}, got {current.type if current else 'EOF'}"
                        raise Exception(
                            msg,
                        )
                # Get function name from expression
                if isinstance(expr, Identifier):
                    expr = FunctionCall(function=expr.name, arguments=args)
                elif isinstance(expr, MemberAccess):
                    # Method call like math.entropy(...)
                    expr = FunctionCall(
                        function=f"{self._get_full_name(expr)}",
                        arguments=args,
                    )
                else:
                    msg = "Invalid function call"
                    raise Exception(msg)
            else:
                break

        return expr

    def _get_full_name(self, expr: Expression) -> str:
        """Get full name from member access chain."""
        if isinstance(expr, Identifier):
            return expr.name
        if isinstance(expr, MemberAccess):
            obj_name = (
                self._get_full_name(expr.object)
                if isinstance(expr.object, Identifier | MemberAccess)
                else str(expr.object)
            )
            return f"{obj_name}.{expr.member}"
        return str(expr)

    def _parse_meta_section(self) -> list[Meta]:
        """Parse meta section."""
        meta_list = []

        while self._current_token() and self._current_token().type == TokenType.IDENTIFIER:
            key = self._current_token().value
            self._advance()

            if not self._match(TokenType.ASSIGN):
                msg = "Expected '=' in meta"
                raise Exception(msg)

            # Parse value
            if self._match(TokenType.STRING):
                value = self._previous().value
            elif self._match(TokenType.MINUS):
                # Handle negative integers
                if not self._match(TokenType.INTEGER):
                    msg = "Expected integer after '-' in meta value"
                    raise Exception(msg)
                value = -self._previous().value
            elif self._match(TokenType.INTEGER):
                value = self._previous().value
            elif self._match(TokenType.BOOLEAN_TRUE):
                value = True
            elif self._match(TokenType.BOOLEAN_FALSE):
                value = False
            else:
                msg = "Expected meta value"
                raise Exception(msg)

            meta_list.append(Meta(key=key, value=value))

        return meta_list

    def _parse_strings_section(self) -> list[StringDefinition]:
        """Parse strings section."""
        strings = []
        anonymous_counter = 0  # Counter for anonymous strings

        while self._current_token() and self._current_token().type == TokenType.STRING_IDENTIFIER:
            identifier = self._current_token().value
            self._advance()

            # Generate unique identifier for anonymous strings
            if identifier == "$":
                anonymous_counter += 1
                identifier = f"$anon_{anonymous_counter}"

            if not self._match(TokenType.ASSIGN):
                msg = "Expected '='"
                raise Exception(msg)

            # Parse string value
            if self._match(TokenType.STRING):
                value = self._previous().value
                # Parse modifiers
                modifiers = []
                while self._current_token() and self._current_token().type in (
                    TokenType.NOCASE,
                    TokenType.WIDE,
                    TokenType.ASCII,
                    TokenType.FULLWORD,
                    TokenType.BASE64,
                    TokenType.BASE64WIDE,
                    TokenType.XOR_MOD,
                ):
                    mod_name = self._current_token().value
                    self._advance()

                    # Handle XOR with optional parameters
                    if (
                        mod_name.lower() == "xor"
                        and self._current_token()
                        and self._current_token().type == TokenType.LPAREN
                    ):
                        self._advance()  # consume '('
                        # Parse XOR parameter (can be single value or range)
                        # For now, just skip to closing paren
                        depth = 1
                        while depth > 0 and self._current_token():
                            if self._current_token().type == TokenType.LPAREN:
                                depth += 1
                            elif self._current_token().type == TokenType.RPAREN:
                                depth -= 1
                            self._advance()
                        modifiers.append(StringModifier(name="xor"))
                    else:
                        modifiers.append(StringModifier(name=mod_name))
                strings.append(
                    PlainString(identifier=identifier, value=value, modifiers=modifiers),
                )
            elif self._match(TokenType.HEX_STRING):
                # Parse hex string content
                hex_content = self._previous().value.strip()
                hex_tokens = []

                # Parse hex string content including wildcards
                from yaraast.ast.strings import HexByte, HexWildcard

                # Remove any whitespace and parse pairs of hex digits
                hex_clean = hex_content.replace(" ", "").replace("\t", "").replace("\n", "")

                i = 0
                while i < len(hex_clean):
                    if i + 1 < len(hex_clean):
                        two_chars = hex_clean[i : i + 2]
                        if two_chars == "??":
                            # Wildcard
                            hex_tokens.append(HexWildcard())
                            i += 2
                        elif all(c in "0123456789ABCDEFabcdef" for c in two_chars):
                            # Hex byte - convert to integer
                            hex_tokens.append(HexByte(value=int(two_chars, 16)))
                            i += 2
                        else:
                            # Skip invalid character
                            i += 1
                    else:
                        # Skip single character at end
                        i += 1

                strings.append(HexString(identifier=identifier, tokens=hex_tokens))
            elif self._match(TokenType.REGEX):
                # Regex string
                regex_val = self._previous().value
                # Extract modifiers if present
                modifiers = []
                pattern = regex_val

                # The lexer returns pattern or pattern\x00modifiers
                if "\x00" in regex_val:
                    parts = regex_val.split("\x00", 1)
                    pattern = parts[0]
                    mod_str = parts[1] if len(parts) > 1 else ""
                    for m in mod_str:
                        if m == "i":
                            modifiers.append(StringModifier(name="nocase"))
                        elif m == "s":
                            modifiers.append(StringModifier(name="dotall"))

                # Parse YARA modifiers (nocase, wide, etc.) after the regex
                while self._current_token() and self._current_token().type in (
                    TokenType.NOCASE,
                    TokenType.WIDE,
                    TokenType.ASCII,
                    TokenType.FULLWORD,
                    TokenType.BASE64,
                    TokenType.BASE64WIDE,
                    TokenType.XOR_MOD,
                ):
                    mod_name = self._current_token().value
                    self._advance()

                    # Handle XOR with optional parameters
                    if (
                        mod_name.lower() == "xor"
                        and self._current_token()
                        and self._current_token().type == TokenType.LPAREN
                    ):
                        self._advance()  # consume '('
                        # Parse XOR parameter (can be single value or range)
                        # For now, just skip to closing paren
                        depth = 1
                        while depth > 0 and self._current_token():
                            if self._current_token().type == TokenType.LPAREN:
                                depth += 1
                            elif self._current_token().type == TokenType.RPAREN:
                                depth -= 1
                            self._advance()
                        modifiers.append(StringModifier(name="xor"))
                    else:
                        modifiers.append(StringModifier(name=mod_name))

                strings.append(
                    RegexString(
                        identifier=identifier,
                        regex=pattern,
                        modifiers=modifiers,
                    ),
                )
            else:
                msg = "Expected string value"
                raise Exception(msg)

        return strings

    def _parse_import(self) -> Import:
        """Parse import statement."""
        if not self._match(TokenType.STRING):
            msg = "Expected module name"
            raise Exception(msg)

        module = self._previous().value
        alias = None

        # Check for 'as alias'
        if self._match(TokenType.AS):
            if not self._match(TokenType.IDENTIFIER):
                msg = "Expected alias after 'as'"
                raise Exception(msg)
            alias = self._previous().value

        return Import(module=module, alias=alias)

    def _parse_include(self) -> Include:
        """Parse include statement."""
        if not self._match(TokenType.STRING):
            msg = "Expected file path"
            raise Exception(msg)

        return Include(path=self._previous().value)

    # Helper methods
    def _current_token(self) -> Token | None:
        """Get current token."""
        if self.position < len(self.tokens):
            return self.tokens[self.position]
        return None

    def _previous(self) -> Token:
        """Get previous token."""
        return self.tokens[self.position - 1]

    def _advance(self) -> None:
        """Advance to next token."""
        if self.position < len(self.tokens):
            self.position += 1

    def _parse_for_expression(self) -> Expression:
        """Parse for expression."""
        # Parse quantifier (any, all, or number)
        if self._match(TokenType.ANY):
            quantifier = "any"
        elif self._match(TokenType.ALL):
            quantifier = "all"
        elif self._match(TokenType.INTEGER):
            quantifier = self._previous().value
        else:
            msg = "Expected quantifier after 'for'"
            raise Exception(msg)

        # Check for 'of' (for...of expression)
        if self._match(TokenType.OF):
            # for...of expression
            # Parse string set which could be:
            # - them
            # - ($a, $b, $c)
            # - ($a*)
            # - $a
            if self._match(TokenType.THEM):
                string_set = Identifier(name="them")
            elif self._match(TokenType.LPAREN):
                # Parse string set like ($a, $b, $c) or ($a*)
                string_ids = []
                while not self._check(TokenType.RPAREN) and not self._is_at_end():
                    if self._match(TokenType.STRING_IDENTIFIER):
                        string_name = self._previous().value
                        # Check for wildcard pattern (e.g., $a*)
                        if self._match(TokenType.MULTIPLY):
                            string_name += "*"
                        from yaraast.ast.expressions import StringIdentifier

                        string_ids.append(StringIdentifier(name=string_name))
                    elif self._match(TokenType.MULTIPLY):
                        # Handle standalone wildcards (*)
                        from yaraast.ast.expressions import StringIdentifier

                        string_ids.append(StringIdentifier(name="*"))
                    else:
                        # Skip unexpected tokens and try to continue
                        if self._current_token():
                            self._advance()
                        continue

                    if not self._match(TokenType.COMMA):
                        break

                if not self._match(TokenType.RPAREN):
                    # Try to recover by finding the closing parenthesis
                    while not self._is_at_end() and not self._check(TokenType.RPAREN):
                        self._advance()
                    if not self._match(TokenType.RPAREN):
                        msg = "Expected ')' after string set"
                        raise Exception(msg)

                from yaraast.ast.expressions import SetExpression

                string_set = SetExpression(elements=string_ids)
            else:
                # Single string or identifier
                string_set = self._parse_primary_expression()

            # Optional condition
            body = None
            if self._match(TokenType.COLON) and self._match(TokenType.LPAREN):
                # Save position in case we need to retry
                saved_pos = self.position
                try:
                    body = self._parse_expression()
                except (ValueError, TypeError, AttributeError):
                    # Reset and try to parse with special handling for $
                    self.position = saved_pos

                    # Check if it starts with $ (anonymous string reference in for context)
                    if (
                        self._current_token()
                        and self._current_token().type == TokenType.STRING_IDENTIFIER
                    ):
                        token_val = self._current_token().value
                        if token_val == "$":
                            # This is a reference to the current string in the for loop
                            self._advance()
                            from yaraast.ast.expressions import StringIdentifier

                            current_string = StringIdentifier(name="$")

                            # Now parse the rest (like "in (0..65536)")
                            if self._match(TokenType.IN):
                                range_expr = self._parse_additive_expression()
                                from yaraast.ast.conditions import InExpression

                                body = InExpression(string_id="$", range=range_expr)
                            else:
                                body = current_string
                        else:
                            # Normal string identifier, retry parsing
                            body = self._parse_expression()
                    else:
                        # If all else fails, use default
                        body = BooleanLiteral(value=True)
                        while not self._is_at_end() and not self._check(
                            TokenType.RPAREN,
                        ):
                            self._advance()

                if not self._match(TokenType.RPAREN):
                    # Try to recover by finding the closing parenthesis
                    current = self._current_token()
                    attempts = 0
                    while not self._is_at_end() and attempts < 100:
                        if self._match(TokenType.RPAREN):
                            break
                        self._advance()
                        attempts += 1
                    else:
                        msg = f"Expected ')' after for...of condition at position {self.position}, got {current.type if current else 'EOF'}"
                        raise Exception(
                            msg,
                        )

            # Create ForOfExpression
            from yaraast.ast.conditions import ForOfExpression
            from yaraast.ast.expressions import IntegerLiteral, StringLiteral

            # Convert quantifier to appropriate Expression type
            if isinstance(quantifier, int):
                quantifier_expr = IntegerLiteral(value=str(quantifier))
            elif isinstance(quantifier, str):
                if quantifier in ["any", "all"]:
                    quantifier_expr = Identifier(name=quantifier)
                else:
                    quantifier_expr = StringLiteral(value=quantifier)
            else:
                quantifier_expr = quantifier

            return ForOfExpression(
                quantifier=quantifier_expr,
                string_set=string_set,
                condition=body,
            )

        # Otherwise it's a regular for expression (for all i in (0..10) : (...))
        if not self._match(TokenType.IDENTIFIER):
            msg = "Expected variable name after quantifier"
            raise Exception(msg)

        variable = self._previous().value

        if not self._match(TokenType.IN):
            msg = "Expected 'in' after variable"
            raise Exception(msg)

        iterable = self._parse_primary_expression()

        if not self._match(TokenType.COLON):
            msg = "Expected ':' after iterable"
            raise Exception(msg)

        if not self._match(TokenType.LPAREN):
            msg = "Expected '(' after ':'"
            raise Exception(msg)

        try:
            body = self._parse_expression()
        except (ValueError, TypeError, AttributeError):
            # If expression parsing fails, try to recover
            body = BooleanLiteral(value=True)
            while not self._is_at_end() and not self._check(TokenType.RPAREN):
                self._advance()

        if not self._match(TokenType.RPAREN):
            # Try to recover by finding the closing parenthesis
            current = self._current_token()
            attempts = 0
            while not self._is_at_end() and attempts < 100:
                if self._match(TokenType.RPAREN):
                    break
                self._advance()
                attempts += 1
            else:
                msg = f"Expected ')' after for body at position {self.position}, got {current.type if current else 'EOF'}"
                raise Exception(
                    msg,
                )

        from yaraast.ast.conditions import ForExpression

        return ForExpression(
            quantifier=quantifier,
            variable=variable,
            iterable=iterable,
            body=body,
        )

    def _parse_of_expression(self, quantifier) -> Expression:
        """Parse of expression (e.g., '2 of them')."""
        # Convert quantifier to appropriate Expression type
        from yaraast.ast.expressions import IntegerLiteral, StringLiteral

        if isinstance(quantifier, int):
            quantifier_expr = IntegerLiteral(value=str(quantifier))
        elif isinstance(quantifier, str):
            if quantifier in ["any", "all"]:
                quantifier_expr = Identifier(name=quantifier)
            else:
                quantifier_expr = StringLiteral(value=quantifier)
        else:
            quantifier_expr = quantifier

        # Parse the string set expression
        if self._match(TokenType.THEM):
            string_set = Identifier(name="them")
        elif self._match(TokenType.LPAREN):
            # Parse string set like ($a, $b, $c)
            string_ids = []
            while not self._check(TokenType.RPAREN) and not self._is_at_end():
                if self._match(TokenType.STRING_IDENTIFIER):
                    string_name = self._previous().value
                    # Check for wildcard pattern (e.g., $a*)
                    if self._match(TokenType.MULTIPLY):
                        string_name += "*"
                    string_ids.append(StringIdentifier(name=string_name))
                elif self._match(TokenType.MULTIPLY):
                    # Handle standalone wildcards (*)
                    string_ids.append(StringIdentifier(name="*"))
                elif self._match(TokenType.INTEGER):
                    # Handle integer literals in set expressions
                    value = self._previous().value
                    string_ids.append(IntegerLiteral(value=value))
                else:
                    # Skip unexpected tokens and try to continue
                    # This helps with parsing files with unsupported syntax
                    if self._current_token():
                        self._advance()
                    continue

                if not self._match(TokenType.COMMA):
                    break

            if not self._match(TokenType.RPAREN):
                # Try to recover by finding the closing parenthesis
                while not self._is_at_end() and not self._check(TokenType.RPAREN):
                    self._advance()
                if not self._match(TokenType.RPAREN):
                    msg = "Expected ')' after string set"
                    raise Exception(msg)

            from yaraast.ast.expressions import SetExpression

            string_set = SetExpression(elements=string_ids)
        else:
            # Try to parse a primary expression
            string_set = self._parse_primary_expression()

        from yaraast.ast.conditions import OfExpression

        return OfExpression(quantifier=quantifier_expr, string_set=string_set)

    def _parse_expression(self) -> Expression:
        """Parse a general expression."""
        return self._parse_or_expression()

    def _match(self, *types: TokenType) -> bool:
        """Check if current token matches any of the given types."""
        for token_type in types:
            if self._check(token_type):
                self._advance()
                return True
        return False

    def _check(self, token_type: TokenType) -> bool:
        """Check if current token is of given type."""
        if self._is_at_end():
            return False
        token = self._current_token()
        return token and token.type == token_type

    def _is_at_end(self) -> bool:
        """Check if at end of tokens."""
        return self.position >= len(self.tokens) or (
            self._current_token() and self._current_token().type == TokenType.EOF
        )



"""String-related AST nodes."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any

from yaraast.ast.base import ASTNode
from yaraast.ast.modifiers import StringModifier as EnhancedStringModifier


# Keep backward compatibility with old StringModifier
@dataclass
class StringModifier(ASTNode):
    """Legacy string modifier for backward compatibility."""

    name: str
    value: Any | None = None

    def accept(self, visitor: Any) -> Any:
        return visitor.visit_string_modifier(self)

    def to_legacy_modifier(self) -> StringModifier:
        """Return self as already a legacy modifier."""
        return self

    def to_enhanced_modifier(self) -> EnhancedStringModifier:
        """Convert to enhanced enum-based modifier."""
        return EnhancedStringModifier.from_name_value(self.name, self.value)


@dataclass
class StringDefinition(ASTNode):
    """Base class for string definitions."""

    identifier: str
    modifiers: list[StringModifier] = field(default_factory=list)

    def accept(self, visitor: Any) -> Any:
        return visitor.visit_string_definition(self)


@dataclass
class PlainString(StringDefinition):
    """Plain text string definition."""

    value: str = ""  # Add default

    def accept(self, visitor: Any) -> Any:
        return visitor.visit_plain_string(self)


@dataclass
class HexString(StringDefinition):
    """Hex string definition."""

    tokens: list[HexToken] = field(default_factory=list)  # Add default

    def accept(self, visitor: Any) -> Any:
        return visitor.visit_hex_string(self)


@dataclass
class HexToken(ASTNode):
    """Base class for hex string tokens."""

    def accept(self, visitor: Any) -> Any:
        return visitor.visit_hex_token(self)


@dataclass
class HexByte(HexToken):
    """Single hex byte."""

    value: int = 0  # Add default

    def accept(self, visitor: Any) -> Any:
        return visitor.visit_hex_byte(self)


@dataclass
class HexWildcard(HexToken):
    """Hex wildcard (?)."""

    def accept(self, visitor: Any) -> Any:
        return visitor.visit_hex_wildcard(self)


@dataclass
class HexJump(HexToken):
    """Hex jump [n-m]."""

    min_jump: int | None = None
    max_jump: int | None = None

    def accept(self, visitor: Any) -> Any:
        return visitor.visit_hex_jump(self)


@dataclass
class HexAlternative(HexToken):
    """Hex alternative (a|b|c)."""

    alternatives: list[list[HexToken]] = field(default_factory=list)  # Add default

    def accept(self, visitor: Any) -> Any:
        return visitor.visit_hex_alternative(self)


@dataclass
class HexNibble(HexToken):
    """Hex nibble (half-byte) pattern."""

    high: bool  # True for X?, False for ?X
    value: int = 0  # 0-15

    def accept(self, visitor: Any) -> Any:
        return visitor.visit_hex_nibble(self)


@dataclass
class RegexString(StringDefinition):
    """Regular expression string."""

    regex: str = ""  # Add default

    def accept(self, visitor: Any) -> Any:
        return visitor.visit_regex_string(self)



"""YARA-L specific AST nodes."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any

from yaraast.ast.base import ASTNode


@dataclass
class YaraLRule(ASTNode):
    """YARA-L rule AST node."""

    name: str
    meta: MetaSection | None = None
    events: EventsSection | None = None
    match: MatchSection | None = None
    condition: ConditionSection | None = None
    outcome: OutcomeSection | None = None
    options: OptionsSection | None = None

    def accept(self, visitor: Any) -> Any:
        return visitor.visit_yaral_rule(self)


@dataclass
class MetaSection(ASTNode):
    """YARA-L meta section."""

    entries: list[MetaEntry] = field(default_factory=list)

    def accept(self, visitor: Any) -> Any:
        return visitor.visit_yaral_meta_section(self)


@dataclass
class MetaEntry(ASTNode):
    """Meta entry (key-value pair)."""

    key: str
    value: str | int | bool

    def accept(self, visitor: Any) -> Any:
        return visitor.visit_yaral_meta_entry(self)


@dataclass
class EventsSection(ASTNode):
    """YARA-L events section."""

    statements: list[EventStatement] = field(default_factory=list)

    def accept(self, visitor: Any) -> Any:
        return visitor.visit_yaral_events_section(self)


@dataclass
class EventStatement(ASTNode):
    """Single event statement."""

    def accept(self, visitor: Any) -> Any:
        return visitor.visit_yaral_event_statement(self)


@dataclass
class EventAssignment(EventStatement):
    """Event field assignment: $e.field = value."""

    event_var: EventVariable
    field_path: UDMFieldPath
    operator: str  # =, !=, >, <, >=, <=, in, regex, etc.
    value: str | int | EventVariable | UDMFieldPath | ReferenceList
    modifiers: list[str] = field(default_factory=list)  # nocase, etc.

    def accept(self, visitor: Any) -> Any:
        return visitor.visit_yaral_event_assignment(self)


@dataclass
class EventVariable(ASTNode):
    """Event variable like $e, $e1, $login."""

    name: str

    def accept(self, visitor: Any) -> Any:
        return visitor.visit_yaral_event_variable(self)


@dataclass
class UDMFieldPath(ASTNode):
    """UDM field path like metadata.event_type or principal.hostname."""

    parts: list[str]

    @property
    def path(self) -> str:
        return ".".join(self.parts)

    def accept(self, visitor: Any) -> Any:
        return visitor.visit_yaral_udm_field_path(self)


@dataclass
class UDMFieldAccess(ASTNode):
    """Access to UDM field with event variable."""

    event: EventVariable
    field: UDMFieldPath

    @property
    def full_path(self) -> str:
        return f"{self.event.name}.{self.field.path}"

    def accept(self, visitor: Any) -> Any:
        return visitor.visit_yaral_udm_field_access(self)


@dataclass
class ReferenceList(ASTNode):
    """Reference list like %suspicious_ips%."""

    name: str

    def accept(self, visitor: Any) -> Any:
        return visitor.visit_yaral_reference_list(self)


@dataclass
class MatchSection(ASTNode):
    """YARA-L match section for time windows."""

    variables: list[MatchVariable] = field(default_factory=list)

    def accept(self, visitor: Any) -> Any:
        return visitor.visit_yaral_match_section(self)


@dataclass
class MatchVariable(ASTNode):
    """Match variable with time window."""

    variable: str  # Variable name (without $)
    time_window: TimeWindow

    def accept(self, visitor: Any) -> Any:
        return visitor.visit_yaral_match_variable(self)


@dataclass
class TimeWindow(ASTNode):
    """Time window specification."""

    duration: int
    unit: str  # s, m, h, d
    modifier: str | None = None  # 'every' for grouped windows

    @property
    def as_string(self) -> str:
        prefix = f"{self.modifier} " if self.modifier else ""
        return f"{prefix}{self.duration}{self.unit}"

    def accept(self, visitor: Any) -> Any:
        return visitor.visit_yaral_time_window(self)


@dataclass
class ConditionSection(ASTNode):
    """YARA-L condition section."""

    expression: ConditionExpression

    def accept(self, visitor: Any) -> Any:
        return visitor.visit_yaral_condition_section(self)


@dataclass
class ConditionExpression(ASTNode):
    """Base class for condition expressions."""

    def accept(self, visitor: Any) -> Any:
        return visitor.visit_yaral_condition_expression(self)


@dataclass
class BinaryCondition(ConditionExpression):
    """Binary condition (AND, OR)."""

    operator: str  # and, or
    left: ConditionExpression
    right: ConditionExpression

    def accept(self, visitor: Any) -> Any:
        return visitor.visit_yaral_binary_condition(self)


@dataclass
class UnaryCondition(ConditionExpression):
    """Unary condition (NOT)."""

    operator: str  # not
    operand: ConditionExpression

    def accept(self, visitor: Any) -> Any:
        return visitor.visit_yaral_unary_condition(self)


@dataclass
class EventCountCondition(ConditionExpression):
    """Event count condition like #e > 5."""

    event: str  # Event variable name (without $)
    operator: str  # >, <, >=, <=, ==, !=
    count: int

    def accept(self, visitor: Any) -> Any:
        return visitor.visit_yaral_event_count_condition(self)


@dataclass
class EventExistsCondition(ConditionExpression):
    """Check if event exists: $e1."""

    event: str  # Event variable name

    def accept(self, visitor: Any) -> Any:
        return visitor.visit_yaral_event_exists_condition(self)


@dataclass
class VariableComparisonCondition(ConditionExpression):
    """Variable comparison condition like $var > 5 or $dc_count > 3."""

    variable: str  # Variable name (with or without $)
    operator: str  # >, <, >=, <=, ==, !=
    value: Any  # Comparison value

    def accept(self, visitor: Any) -> Any:
        return visitor.visit_yaral_variable_comparison_condition(self)


@dataclass
class JoinCondition(ConditionExpression):
    """Join condition between events."""

    left_event: str
    right_event: str
    join_type: str = "inner"  # inner, left, right, full

    def accept(self, visitor: Any) -> Any:
        return visitor.visit_yaral_join_condition(self)


@dataclass
class OutcomeSection(ASTNode):
    """YARA-L outcome section for extracting data."""

    assignments: list[OutcomeAssignment] = field(default_factory=list)

    def accept(self, visitor: Any) -> Any:
        return visitor.visit_yaral_outcome_section(self)


@dataclass
class OutcomeAssignment(ASTNode):
    """Outcome variable assignment."""

    variable: str  # Variable name (with $)
    expression: OutcomeExpression

    def accept(self, visitor: Any) -> Any:
        return visitor.visit_yaral_outcome_assignment(self)


@dataclass
class OutcomeExpression(ASTNode):
    """Base class for outcome expressions."""

    def accept(self, visitor: Any) -> Any:
        return visitor.visit_yaral_outcome_expression(self)


@dataclass
class AggregationFunction(OutcomeExpression):
    """Aggregation function call."""

    function: str  # count, count_distinct, sum, min, max, array, etc.
    arguments: list[UDMFieldAccess | str]

    @property
    def call_string(self) -> str:
        args = ", ".join(str(arg) for arg in self.arguments)
        return f"{self.function}({args})"

    def accept(self, visitor: Any) -> Any:
        return visitor.visit_yaral_aggregation_function(self)


@dataclass
class ConditionalExpression(OutcomeExpression):
    """Conditional expression: if(condition, true_value, false_value)."""

    condition: Any  # Can be various condition types
    true_value: Any
    false_value: Any

    def accept(self, visitor: Any) -> Any:
        return visitor.visit_yaral_conditional_expression(self)


@dataclass
class ArithmeticExpression(OutcomeExpression):
    """Arithmetic expression."""

    operator: str  # +, -, *, /
    left: Any
    right: Any

    def accept(self, visitor: Any) -> Any:
        return visitor.visit_yaral_arithmetic_expression(self)


@dataclass
class OptionsSection(ASTNode):
    """YARA-L options section."""

    options: dict[str, Any] = field(default_factory=dict)

    def accept(self, visitor: Any) -> Any:
        return visitor.visit_yaral_options_section(self)


@dataclass
class RegexPattern(ASTNode):
    """Regex pattern with optional flags."""

    pattern: str
    flags: list[str] = field(default_factory=list)  # nocase, etc.

    @property
    def as_string(self) -> str:
        flags_str = " ".join(self.flags) if self.flags else ""
        return f"/{self.pattern}/ {flags_str}".strip()

    def accept(self, visitor: Any) -> Any:
        return visitor.visit_yaral_regex_pattern(self)


@dataclass
class CIDRExpression(ASTNode):
    """CIDR expression for IP matching."""

    field: UDMFieldAccess
    cidr: str  # e.g., "192.168.1.0/24"

    def accept(self, visitor: Any) -> Any:
        return visitor.visit_yaral_cidr_expression(self)


@dataclass
class FunctionCall(ASTNode):
    """Function call like re.regex() or cidr()."""

    function: str
    arguments: list[Any]

    @property
    def call_string(self) -> str:
        args = ", ".join(str(arg) for arg in self.arguments)
        return f"{self.function}({args})"

    def accept(self, visitor: Any) -> Any:
        return visitor.visit_yaral_function_call(self)


@dataclass
class YaraLFile(ASTNode):
    """YARA-L file containing multiple rules."""

    rules: list[YaraLRule] = field(default_factory=list)

    def add_rule(self, rule: YaraLRule) -> None:
        """Add a rule to the file."""
        self.rules.append(rule)

    def accept(self, visitor: Any) -> Any:
        return visitor.visit_yaral_file(self)



"""Test module loader functionality."""

import json
import os
import tempfile
from pathlib import Path

from yaraast.types.module_loader import ModuleLoader
from yaraast.types.type_system import ArrayType, IntegerType


def test_builtin_modules() -> None:
    """Test loading of builtin modules."""
    loader = ModuleLoader()

    # Should have builtin modules
    assert "pe" in loader.modules
    assert "elf" in loader.modules
    assert "math" in loader.modules

    # Check PE module
    pe_module = loader.get_module("pe")
    assert pe_module is not None
    assert "machine" in pe_module.attributes
    assert "imphash" in pe_module.functions


def test_load_json_module() -> None:
    """Test loading module from JSON file."""
    # Create temporary JSON module
    module_json = {
        "name": "test_module",
        "attributes": {"version": "int", "data": "string[]"},
        "functions": {
            "process": {
                "return": "bool",
                "parameters": [{"name": "input", "type": "string"}],
            },
        },
        "constants": {"MAX_SIZE": "int"},
    }

    with tempfile.TemporaryDirectory() as tmpdir:
        json_path = Path(tmpdir) / "test_module.json"
        with json_path.open("w") as f:
            json.dump(module_json, f)

        # Set environment variable
        os.environ["YARAAST_MODULE_SPEC_PATH"] = tmpdir

        # Load modules
        loader = ModuleLoader()

        # Check module was loaded
        assert "test_module" in loader.modules
        module = loader.get_module("test_module")
        assert module is not None

        # Check attributes
        assert "version" in module.attributes
        assert isinstance(module.attributes["version"], IntegerType)
        assert "data" in module.attributes
        assert isinstance(module.attributes["data"], ArrayType)

        # Check functions
        assert "process" in module.functions
        func = module.functions["process"]
        assert func.name == "process"
        assert len(func.parameters) == 1
        assert func.parameters[0][0] == "input"

        # Check constants
        assert "MAX_SIZE" in module.constants

        # Clean up
        del os.environ["YARAAST_MODULE_SPEC_PATH"]


def test_exclusive_module_path() -> None:
    """Test exclusive module loading (ignoring builtins)."""
    module_json = [{"name": "custom_only", "attributes": {"test": "string"}}]

    with tempfile.TemporaryDirectory() as tmpdir:
        json_path = Path(tmpdir) / "modules.json"
        with json_path.open("w") as f:
            json.dump(module_json, f)

        # Set exclusive path
        os.environ["YARAAST_MODULE_SPEC_PATH_EXCLUSIVE"] = str(json_path)

        # Load modules
        loader = ModuleLoader()

        # Should only have our custom module
        assert "custom_only" in loader.modules
        assert "pe" not in loader.modules  # Builtin should be excluded
        assert len(loader.modules) == 1

        # Clean up
        del os.environ["YARAAST_MODULE_SPEC_PATH_EXCLUSIVE"]


def test_complex_types() -> None:
    """Test loading complex types from JSON."""
    module_json = {
        "name": "complex",
        "attributes": {
            "simple_array": "int[]",
            "nested": {
                "type": "struct",
                "fields": {"name": "string", "values": "int[]", "enabled": "bool"},
            },
            "dictionary": {"type": "dict", "key": "string", "value": "int"},
        },
    }

    with tempfile.TemporaryDirectory() as tmpdir:
        json_path = Path(tmpdir) / "complex.json"
        with json_path.open("w") as f:
            json.dump(module_json, f)

        os.environ["YARAAST_MODULE_SPEC_PATH"] = tmpdir

        loader = ModuleLoader()
        module = loader.get_module("complex")

        assert module is not None
        assert "simple_array" in module.attributes
        assert "nested" in module.attributes
        assert "dictionary" in module.attributes

        # Clean up
        del os.environ["YARAAST_MODULE_SPEC_PATH"]


def test_example_custom_module() -> None:
    """Test loading the example custom module."""
    # Load from default modules directory
    loader = ModuleLoader()

    # The example_custom.json should be loaded if it exists
    modules_dir = Path(__file__).parent.parent / "yaraast" / "types" / "modules"
    if modules_dir.exists() and (modules_dir / "example_custom.json").exists():
        assert "custom" in loader.modules

        custom = loader.get_module("custom")
        assert custom is not None
        assert "api_version" in custom.attributes
        assert "hash" in custom.functions
        assert "MAX_BUFFER_SIZE" in custom.constants


if __name__ == "__main__":
    test_builtin_modules()
    test_load_json_module()
    test_exclusive_module_path()
    test_complex_types()
    test_example_custom_module()
    print("✓ All module loader tests passed")
